from enum import Enum


class WorkspaceStarterTemplate(str, Enum):
    llm_app = "llm-app"
    api_app = "api-app"
    django_app = "django-app"
    streamlit_app = "streamlit-app"
    ai_platform = "ai-platform"



from phi.api.schemas.ai import ConversationType
from phi.cli.config import PhiCliConfig


def phi_ai_conversation(
    phi_config: PhiCliConfig,
    start_new_conversation: bool = False,
    autonomous_conversation: bool = True,
    print_conversation_history: bool = False,
    stream: bool = False,
) -> None:
    """Start a conversation with Phi AI."""
    from phi.ai.phi_ai import PhiAI

    conversation_type = ConversationType.AUTO if autonomous_conversation else ConversationType.RAG
    ai = PhiAI(new_conversation=start_new_conversation, conversation_type=conversation_type, phi_config=phi_config)
    if print_conversation_history:
        ai.print_conversation_history()

    ai.start_conversation(stream=stream)



from typing import List, Iterator

from phi.document import Document
from phi.knowledge.base import KnowledgeBase


class CombinedKnowledgeBase(KnowledgeBase):
    sources: List[KnowledgeBase] = []

    @property
    def document_lists(self) -> Iterator[List[Document]]:
        """Iterate over knowledge bases and yield lists of documents.
        Each object yielded by the iterator is a list of documents.

        Returns:
            Iterator[List[Document]]: Iterator yielding list of documents
        """

        for kb in self.sources:
            for document_list in kb.document_lists:
                yield document_list



from abc import ABC, abstractmethod
from typing import List

from phi.document import Document


class VectorDb(ABC):
    """Base class for managing Vector Databases"""

    @abstractmethod
    def create(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def doc_exists(self, document: Document) -> bool:
        raise NotImplementedError

    @abstractmethod
    def name_exists(self, name: str) -> bool:
        raise NotImplementedError

    @abstractmethod
    def insert(self, documents: List[Document]) -> None:
        raise NotImplementedError

    @abstractmethod
    def upsert(self, documents: List[Document]) -> None:
        raise NotImplementedError

    @abstractmethod
    def search(self, query: str, limit: int = 5) -> List[Document]:
        raise NotImplementedError

    @abstractmethod
    def delete(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def exists(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def optimize(self) -> None:
        raise NotImplementedError



from typing import List, Optional, Dict, Union

from httpx import Response

from phi.api.api import api, invalid_response
from phi.api.routes import ApiRoutes
from phi.api.schemas.user import UserSchema
from phi.api.schemas.workspace import (
    WorkspaceSchema,
    WorkspaceCreate,
    WorkspaceUpdate,
    WorkspaceDelete,
    WorkspaceEvent,
    UpdatePrimaryWorkspace,
)
from phi.cli.settings import phi_cli_settings
from phi.utils.log import logger


def get_primary_workspace(user: UserSchema) -> Optional[WorkspaceSchema]:
    if not phi_cli_settings.api_enabled:
        return None

    logger.debug("--o-o-- Get primary workspace")
    with api.AuthenticatedClient() as api_client:
        try:
            r: Response = api_client.post(
                ApiRoutes.WORKSPACE_READ_PRIMARY, json=user.model_dump(include={"id_user", "email"})
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            primary_workspace: WorkspaceSchema = WorkspaceSchema.model_validate(response_json)
            if primary_workspace is not None:
                return primary_workspace
        except Exception as e:
            logger.debug(f"Could not get primary workspace: {e}")
    return None


def get_available_workspaces(user: UserSchema) -> Optional[List[WorkspaceSchema]]:
    if not phi_cli_settings.api_enabled:
        return None

    logger.debug("--o-o-- Get available workspaces")
    with api.AuthenticatedClient() as api_client:
        try:
            r: Response = api_client.post(
                ApiRoutes.WORKSPACE_READ_AVAILABLE, json=user.model_dump(include={"id_user", "email"})
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            available_workspaces: List[WorkspaceSchema] = []
            for workspace in response_json:
                if not isinstance(workspace, dict):
                    logger.debug(f"Not a dict: {workspace}")
                    continue
                available_workspaces.append(WorkspaceSchema.model_validate(workspace))
            return available_workspaces
        except Exception as e:
            logger.debug(f"Could not get available workspaces: {e}")
    return None


def create_workspace_for_user(user: UserSchema, workspace: WorkspaceCreate) -> Optional[WorkspaceSchema]:
    if not phi_cli_settings.api_enabled:
        return None

    logger.debug("--o-o-- Create workspace")
    with api.AuthenticatedClient() as api_client:
        try:
            r: Response = api_client.post(
                ApiRoutes.WORKSPACE_CREATE,
                json={
                    "user": user.model_dump(include={"id_user", "email"}),
                    "workspace": workspace.model_dump(exclude_none=True),
                },
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            created_workspace: WorkspaceSchema = WorkspaceSchema.model_validate(response_json)
            if created_workspace is not None:
                return created_workspace
        except Exception as e:
            logger.debug(f"Could not create workspace: {e}")
    return None


def update_workspace_for_user(user: UserSchema, workspace: WorkspaceUpdate) -> Optional[WorkspaceSchema]:
    if not phi_cli_settings.api_enabled:
        return None

    logger.debug("--o-o-- Update workspace")
    with api.AuthenticatedClient() as api_client:
        try:
            r: Response = api_client.post(
                ApiRoutes.WORKSPACE_UPDATE,
                json={
                    "user": user.model_dump(include={"id_user", "email"}),
                    "workspace": workspace.model_dump(exclude_none=True),
                },
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            updated_workspace: WorkspaceSchema = WorkspaceSchema.model_validate(response_json)
            if updated_workspace is not None:
                return updated_workspace
        except Exception as e:
            logger.debug(f"Could not update workspace: {e}")
    return None


def update_primary_workspace_for_user(user: UserSchema, workspace: UpdatePrimaryWorkspace) -> Optional[WorkspaceSchema]:
    if not phi_cli_settings.api_enabled:
        return None

    logger.debug(f"--o-o-- Update primary workspace to: {workspace.ws_name}")
    with api.AuthenticatedClient() as api_client:
        try:
            r: Response = api_client.post(
                ApiRoutes.WORKSPACE_UPDATE_PRIMARY,
                json={
                    "user": user.model_dump(include={"id_user", "email"}),
                    "workspace": workspace.model_dump(exclude_none=True),
                },
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            updated_workspace: WorkspaceSchema = WorkspaceSchema.model_validate(response_json)
            if updated_workspace is not None:
                return updated_workspace
        except Exception as e:
            logger.debug(f"Could not update primary workspace: {e}")
    return None


def delete_workspace_for_user(user: UserSchema, workspace: WorkspaceDelete) -> Optional[WorkspaceSchema]:
    if not phi_cli_settings.api_enabled:
        return None

    logger.debug("--o-o-- Delete workspace")
    with api.AuthenticatedClient() as api_client:
        try:
            r: Response = api_client.post(
                ApiRoutes.WORKSPACE_DELETE,
                json={
                    "user": user.model_dump(include={"id_user", "email"}),
                    "workspace": workspace.model_dump(exclude_none=True),
                },
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            updated_workspace: WorkspaceSchema = WorkspaceSchema.model_validate(response_json)
            if updated_workspace is not None:
                return updated_workspace
        except Exception as e:
            logger.debug(f"Could not delete workspace: {e}")
    return None


def log_workspace_event(user: UserSchema, workspace_event: WorkspaceEvent) -> bool:
    if not phi_cli_settings.api_enabled:
        return False

    logger.debug("--o-o-- Log workspace event")
    with api.AuthenticatedClient() as api_client:
        try:
            r: Response = api_client.post(
                ApiRoutes.WORKSPACE_EVENT_CREATE,
                json={
                    "user": user.model_dump(include={"id_user", "email"}),
                    "event": workspace_event.model_dump(exclude_none=True),
                },
            )
            if invalid_response(r):
                return False

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return False

            if isinstance(response_json, dict) and response_json.get("status") == "success":
                return True
            return False
        except Exception as e:
            logger.debug(f"Could not log workspace wvent: {e}")
    return False



from typing import Optional, Dict, Any, List

from phi.llm.aws.bedrock import AwsBedrock


class AnthropicClaude(AwsBedrock):
    name: str = "AwsBedrockClaude"
    model: str = "anthropic.claude-v2"
    max_tokens: int = 8192
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    top_k: Optional[int] = None
    stop_sequences: Optional[List[str]] = None

    def to_dict(self) -> Dict[str, Any]:
        _dict = super().to_dict()
        _dict["max_tokens"] = self.max_tokens
        _dict["temperature"] = self.temperature
        _dict["top_p"] = self.top_p
        _dict["top_k"] = self.top_k
        _dict["stop_sequences"] = self.stop_sequences
        return _dict

    @property
    def api_kwargs(self) -> Dict[str, Any]:
        kwargs: Dict[str, Any] = {}
        if self.max_tokens:
            kwargs["max_tokens_to_sample"] = self.max_tokens
        if self.temperature:
            kwargs["temperature"] = self.temperature
        if self.top_p:
            kwargs["top_p"] = self.top_p
        if self.top_k:
            kwargs["top_k"] = self.top_k
        if self.stop_sequences:
            kwargs["stop_sequences"] = self.stop_sequences
        return kwargs



from pathlib import Path
from typing import List

from phi.document.base import Document
from phi.document.reader.base import Reader
from phi.utils.log import logger


class PDFReader(Reader):
    """Reader for PDF files"""

    def read(self, path: Path) -> List[Document]:
        if not path:
            raise ValueError("No path provided")

        if not path.exists():
            raise FileNotFoundError(f"Could not find file: {path}")

        try:
            from pypdf import PdfReader as DocumentReader  # noqa: F401
        except ImportError:
            raise ImportError("`pypdf` not installed")

        logger.info(f"Reading: {path}")
        doc_name = path.name.split(".")[0]
        doc_reader = DocumentReader(path)

        documents = [
            Document(
                name=doc_name,
                meta_data={"page": page_number},
                content=page.extract_text(),
            )
            for page_number, page in enumerate(doc_reader.pages, start=1)
        ]
        if self.chunk:
            chunked_documents = []
            for document in documents:
                chunked_documents.extend(self.chunk_document(document))
            return chunked_documents
        return documents


class PDFUrlReader(Reader):
    """Reader for PDF files from URL"""

    def read(self, url: str) -> List[Document]:
        if not url:
            raise ValueError("No url provided")

        from io import BytesIO

        try:
            import httpx
        except ImportError:
            raise ImportError("`httpx` not installed")

        try:
            from pypdf import PdfReader as DocumentReader  # noqa: F401
        except ImportError:
            raise ImportError("`pypdf` not installed")

        logger.info(f"Reading: {url}")
        response = httpx.get(url)

        doc_name = url.split("/")[-1].split(".")[0].replace("/", "_").replace(" ", "_")
        doc_reader = DocumentReader(BytesIO(response.content))

        documents = [
            Document(
                name=doc_name,
                meta_data={"page": page_number},
                content=page.extract_text(),
            )
            for page_number, page in enumerate(doc_reader.pages, start=1)
        ]
        if self.chunk:
            chunked_documents = []
            for document in documents:
                chunked_documents.extend(self.chunk_document(document))
            return chunked_documents
        return documents



from phi.llm.function.registry import FunctionRegistry
from phi.utils.log import logger


class PubMedRegistry(FunctionRegistry):
    def __init__(self):
        super().__init__(name="pubmed_registry")
        self.register(self.get_articles_from_pubmed)

    def get_articles_from_pubmed(self, query: str, num_articles: int = 2) -> str:
        """Gets the abstract for articles related to a query from PubMed: a database of biomedical literature
        Use this function to find information about a medical concept when not available in the knowledge base or Google

        :param query: The query to get related articles for.
        :param num_articles: The number of articles to return.
        :return: JSON string containing the articles
        """
        logger.info(f"Searching Pubmed for: {query}")
        return "Sorry, this capability is not available yet."



from kubernetes.client.models.v1_local_object_reference import V1LocalObjectReference

from phi.k8s.resource.base import K8sObject


class LocalObjectReference(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#localobjectreference-v1-core
    """

    resource_type: str = "LocalObjectReference"

    # Name of the referent. More info: https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#names
    name: str

    def get_k8s_object(self) -> V1LocalObjectReference:
        # Return a V1LocalObjectReference object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_local_object_reference.py
        _v1_local_object_reference = V1LocalObjectReference(name=self.name)
        return _v1_local_object_reference



from typing import List, Optional

from kubernetes.client.models.v1_node_selector import V1NodeSelector
from kubernetes.client.models.v1_node_selector_term import V1NodeSelectorTerm
from kubernetes.client.models.v1_node_selector_requirement import (
    V1NodeSelectorRequirement,
)
from pydantic import Field

from phi.k8s.resource.base import K8sObject


class NodeSelectorRequirement(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#nodeselectorrequirement-v1-core
    """

    resource_type: str = "NodeSelectorRequirement"

    # The label key that the selector applies to.
    key: str
    # Represents a key's relationship to a set of values.
    # Valid operators are In, NotIn, Exists, DoesNotExist. Gt, and Lt.
    # Possible enum values: - `"DoesNotExist"` - `"Exists"` - `"Gt"` - `"In"` - `"Lt"` - `"NotIn"`
    operator: str
    # An array of string values. If the operator is In or NotIn, the values array must be non-empty.
    # If the operator is Exists or DoesNotExist, the values array must be empty.
    # If the operator is Gt or Lt, the values array must have a single element, which will be interpreted as an integer.
    # This array is replaced during a strategic merge patch.
    values: Optional[List[str]]

    def get_k8s_object(
        self,
    ) -> V1NodeSelectorRequirement:
        # Return a V1NodeSelectorRequirement object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_node_selector_requirement.py
        _v1_node_selector_requirement = V1NodeSelectorRequirement(
            key=self.key,
            operator=self.operator,
            values=self.values,
        )
        return _v1_node_selector_requirement


class NodeSelectorTerm(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#nodeselectorterm-v1-core
    """

    resource_type: str = "NodeSelectorTerm"

    # A list of node selector requirements by node's labels.
    match_expressions: Optional[List[NodeSelectorRequirement]] = Field(..., alias="matchExpressions")
    # A list of node selector requirements by node's fields.
    match_fields: Optional[List[NodeSelectorRequirement]] = Field(..., alias="matchFields")

    def get_k8s_object(
        self,
    ) -> V1NodeSelectorTerm:
        # Return a V1NodeSelectorTerm object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_node_selector_term.py
        _v1_node_selector_term = V1NodeSelectorTerm(
            match_expressions=[me.get_k8s_object() for me in self.match_expressions]
            if self.match_expressions
            else None,
            match_fields=[mf.get_k8s_object() for mf in self.match_fields] if self.match_fields else None,
        )
        return _v1_node_selector_term


class NodeSelector(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#nodeselector-v1-core
    """

    resource_type: str = "NodeSelector"

    # A node selector represents the union of the results of one or more label queries over a set of nodes;
    # that is, it represents the OR of the selectors represented by the node selector terms.
    node_selector_terms: List[NodeSelectorTerm] = Field(..., alias="nodeSelectorTerms")

    def get_k8s_object(
        self,
    ) -> V1NodeSelector:
        # Return a V1NodeSelector object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_node_selector.py
        _v1_node_selector = V1NodeSelector(
            node_selector_terms=[nst.get_k8s_object() for nst in self.node_selector_terms]
            if self.node_selector_terms
            else None,
        )
        return _v1_node_selector



from typing import List, Optional

from pydantic import BaseModel

from phi.app.base import AppBase


class AppGroup(BaseModel):
    """AppGroup is a collection of Apps"""

    name: Optional[str] = None
    enabled: bool = True
    apps: Optional[List[AppBase]] = None

    class Config:
        arbitrary_types_allowed = True

    def get_apps(self) -> List[AppBase]:
        if self.enabled and self.apps is not None:
            for app in self.apps:
                if app.group is None and self.name is not None:
                    app.group = self.name
            return self.apps
        return []



from typing import Dict, List, Any, Optional, Tuple

from pydantic import BaseModel

from phi.llm.schemas import Message, References


class ConversationMemory(BaseModel):
    """
    This class provides a memory for a Conversation.
    """

    # Messages between the user and the LLM.
    # Note: the actual prompts are stored in the llm_messages
    chat_history: List[Message] = []
    # Messages(prompts) sent to the LLM and the LLM responses.
    llm_messages: List[Message] = []
    # References from the vector database.
    references: List[References] = []

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump(include={"chat_history", "llm_messages", "references"})

    def add_user_message(self, message: Message) -> None:
        """Adds a message sent by the user to the chat_history."""
        self.chat_history.append(message)

    def add_llm_response(self, message: Message) -> None:
        """Adds the LLM response to the chat_history and llm_messages."""
        self.chat_history.append(message)
        self.llm_messages.append(message)

    def add_system_prompt(self, message: Message) -> None:
        """Adds the system prompt sent to the LLM to llm_messages if this is the first llm message."""
        if len(self.llm_messages) == 0:
            self.llm_messages.append(message)

    def add_chat_message(self, message: Message) -> None:
        """Adds a Message to the chat_history."""
        self.chat_history.append(message)

    def add_llm_message(self, message: Message) -> None:
        """Adds a Message to the llm_messages."""
        self.llm_messages.append(message)

    def add_chat_history(self, messages: List[Message]) -> None:
        """Adds a list of messages to the chat_history."""
        self.chat_history.extend(messages)

    def add_llm_messages(self, messages: List[Message]) -> None:
        """Adds a list of messages to the llm_messages."""
        self.llm_messages.extend(messages)

    def add_references(self, references: References) -> None:
        """Adds references to the references list."""
        self.references.append(references)

    def get_chat_history(self) -> List[Dict[str, Any]]:
        """Returns the chat_history as a list of dictionaries.

        :return: A list of dictionaries representing the chat_history.
        """
        return [message.model_dump(exclude_none=True) for message in self.chat_history]

    def get_last_n_messages(self, last_n: Optional[int] = None) -> List[Message]:
        """Returns the last n messages in the chat_history.

        :param last_n: The number of messages to return from the end of the conversation.
            If None, returns all messages.
        :return: A list of Messages in the chat_history.
        """
        return self.chat_history[-last_n:] if last_n else self.chat_history

    def get_llm_messages(self) -> List[Dict[str, Any]]:
        """Returns the llm_messages as a list of dictionaries."""
        return [message.model_dump(exclude_none=True) for message in self.llm_messages]

    def get_formatted_chat_history(self, num_messages: Optional[int] = None) -> str:
        """Returns the chat_history as a formatted string."""

        messages = self.get_last_n_messages(num_messages)
        if len(messages) == 0:
            return ""

        history = ""
        for message in self.get_last_n_messages(num_messages):
            if message.role == "user":
                history += "\n---\n"
            history += f"{message.role.upper()}: {message.content}\n"
        return history

    def get_chats(self) -> List[Tuple[Message, Message]]:
        """Returns a list of tuples of user messages and LLM responses."""

        all_chats: List[Tuple[Message, Message]] = []
        current_chat: List[Message] = []

        # Make a copy of the chat_history and remove all system messages from the beginning.
        chat_history = self.chat_history.copy()
        while len(chat_history) > 0 and chat_history[0].role in ("system", "assistant"):
            chat_history = chat_history[1:]

        for m in chat_history:
            if m.role == "system":
                continue
            if m.role == "user":
                # This is a new chat record
                if len(current_chat) == 2:
                    all_chats.append((current_chat[0], current_chat[1]))
                    current_chat = []
                current_chat.append(m)
            if m.role == "assistant":
                current_chat.append(m)

        if len(current_chat) >= 1:
            all_chats.append((current_chat[0], current_chat[1]))
        return all_chats



from dataclasses import dataclass


@dataclass
class ApiRoutes:
    # user paths
    USER_HEALTH: str = "/v1/user/health"
    USER_READ: str = "/v1/user/read"
    USER_CREATE: str = "/v1/user/create"
    USER_UPDATE: str = "/v1/user/update"
    USER_SIGN_IN: str = "/v1/user/signin"
    USER_CLI_AUTH: str = "/v1/user/cliauth"
    USER_AUTHENTICATE: str = "/v1/user/authenticate"
    USER_AUTH_REFRESH: str = "/v1/user/authrefresh"

    # workspace paths
    WORKSPACE_HEALTH: str = "/v1/workspace/health"
    WORKSPACE_CREATE: str = "/v1/workspace/create"
    WORKSPACE_UPDATE: str = "/v1/workspace/update"
    WORKSPACE_DELETE: str = "/v1/workspace/delete"
    WORKSPACE_EVENT_CREATE: str = "/v1/workspace/event/create"
    WORKSPACE_UPDATE_PRIMARY: str = "/v1/workspace/update/primary"
    WORKSPACE_READ_PRIMARY: str = "/v1/workspace/read/primary"
    WORKSPACE_READ_AVAILABLE: str = "/v1/workspace/read/available"

    # monitor paths
    MONITOR_EVENT_CREATE: str = "/v1/monitor/event/create"

    # conversation paths
    CONVERSATION_UPSERT: str = "/v1/conversation/upsert"
    CONVERSATION_EVENT_CREATE: str = "/v1/conversation/event/create"

    # ai paths
    AI_CONVERSATION_CREATE: str = "/v1/ai/conversation/create"
    AI_CONVERSATION_CHAT: str = "/v1/ai/conversation/chat"
    AI_CONVERSATION_CHAT_WS: str = "/v1/ai/conversation/chat_ws"

    # llm paths
    OPENAI_CHAT: str = "/v1/llm/openai/chat"
    OPENAI_EMBEDDING: str = "/v1/llm/openai/embedding"



from typing import Optional, Dict, Iterator, Any

from httpx import Response

from phi.api.api import api, invalid_response
from phi.api.routes import ApiRoutes
from phi.constants import WORKSPACE_ID_ENV_VAR, WORKSPACE_HASH_ENV_VAR
from phi.utils.env import get_from_env
from phi.utils.log import logger


def openai_chat(params: Dict[str, Any]) -> Optional[Dict]:
    with api.AuthenticatedClient() as api_client:
        logger.debug("--o-o-- OpenAI Chat")
        try:
            workspace_id = get_from_env(WORKSPACE_ID_ENV_VAR)
            workspace_hash = get_from_env(WORKSPACE_HASH_ENV_VAR)

            resp: Response = api_client.post(
                ApiRoutes.OPENAI_CHAT,
                json={
                    "workspace": {
                        "id_workspace": workspace_id,
                        "ws_hash": workspace_hash,
                    },
                    "params": params,
                },
            )
            if invalid_response(resp):
                return None

            return resp.json()
        except Exception as e:
            logger.error(f"Error: {e}")
            logger.info("Please message us on https://discord.gg/4MtYHHrgA8 for help.")
            exit(1)


def openai_chat_stream(params: Dict[str, Any]) -> Iterator[str]:
    with api.AuthenticatedClient() as api_client:
        logger.debug("--o-o-- OpenAI Chat Stream")
        try:
            workspace_id = get_from_env(WORKSPACE_ID_ENV_VAR)
            workspace_hash = get_from_env(WORKSPACE_HASH_ENV_VAR)

            with api_client.stream(
                "POST",
                ApiRoutes.OPENAI_CHAT,
                json={
                    "workspace": {
                        "id_workspace": workspace_id,
                        "ws_hash": workspace_hash,
                    },
                    "params": params,
                },
            ) as streaming_resp:
                for chunk in streaming_resp.iter_text():
                    yield chunk
        except Exception as e:
            logger.error(f"Error: {e}")
            logger.info("Please message us on https://discord.gg/4MtYHHrgA8 for help.")
            exit(1)


def openai_embedding(params: Dict[str, Any]) -> Optional[Dict]:
    with api.AuthenticatedClient() as api_client:
        logger.debug("--o-o-- OpenAI Embedding")
        try:
            workspace_id = get_from_env(WORKSPACE_ID_ENV_VAR)
            workspace_hash = get_from_env(WORKSPACE_HASH_ENV_VAR)

            resp: Response = api_client.post(
                ApiRoutes.OPENAI_EMBEDDING,
                json={
                    "workspace": {
                        "id_workspace": workspace_id,
                        "ws_hash": workspace_hash,
                    },
                    "params": params,
                },
            )
            if invalid_response(resp):
                return None

            return resp.json()
        except Exception as e:
            logger.error(f"Error: {e}")
            logger.info("Please message us on https://discord.gg/4MtYHHrgA8 for help.")
            exit(1)



from phi.k8s.app.postgres.postgres import (
    PostgresDb,
    AppVolumeType,
    ContainerContext,
    ServiceType,
    RestartPolicy,
    ImagePullPolicy,
)

from phi.k8s.app.postgres.pgvector import PgVectorDb



from typing import Optional, Union, Dict, List

from httpx import Response

from phi.api.api import api, invalid_response
from phi.api.routes import ApiRoutes
from phi.api.schemas.monitor import MonitorEventSchema, MonitorResponseSchema
from phi.api.schemas.workspace import WorkspaceSchema
from phi.cli.settings import phi_cli_settings
from phi.utils.log import logger


def log_monitor_event(monitor: MonitorEventSchema, workspace: WorkspaceSchema) -> Optional[MonitorResponseSchema]:
    if not phi_cli_settings.api_enabled:
        return None

    logger.debug("--o-o-- Log monitor event")
    with api.Client() as api_client:
        try:
            r: Response = api_client.post(
                ApiRoutes.MONITOR_EVENT_CREATE,
                json={
                    "monitor": monitor.model_dump(exclude_none=True),
                    "workspace": workspace.model_dump(include={"id_workspace"}),
                },
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            monitor_response: MonitorResponseSchema = MonitorResponseSchema.model_validate(response_json)
            if monitor_response is not None:
                return monitor_response
        except Exception as e:
            logger.debug(f"Could not log monitor event: {e}")
    return None



from phi.docker.app.superset.base import SupersetBase, ContainerContext
from phi.docker.app.superset.webserver import SupersetWebserver
from phi.docker.app.superset.worker import SupersetWorker
from phi.docker.app.superset.worker_beat import SupersetWorkerBeat
from phi.docker.app.superset.init import SupersetInit



from typing import Optional, Dict

from pydantic import BaseModel


class K8sBuildContext(BaseModel):
    namespace: str = "default"
    context: Optional[str] = None
    service_account_name: Optional[str] = None
    labels: Optional[Dict[str, str]] = None



from typing import Optional, Dict, List, Tuple

from phi.embedder.base import Embedder
from phi.utils.env import get_from_env
from phi.utils.log import logger

try:
    from openai import Embedding  # noqa: F401
except ImportError:
    raise ImportError("`openai` not installed")


class OpenAIEmbedder(Embedder):
    model: str = "text-embedding-ada-002"
    dimensions: int = 1536

    def _response(self, text: str):
        if get_from_env("OPENAI_API_KEY") is None:
            logger.debug("--o-o-- Using Phidata Servers")
            try:
                from phi.api.llm import openai_embedding

                response_dict = openai_embedding(
                    params={
                        "input": text,
                        "model": self.model,
                    }
                )
                return response_dict
            except Exception as e:
                logger.exception(e)
                logger.info("Please message us on https://discord.gg/4MtYHHrgA8 for help.")
                exit(1)
        else:
            return Embedding.create(input=text, model=self.model)

    def get_embedding(self, text: str) -> List[float]:
        response = self._response(text=text)
        if "data" not in response:
            return []

        return response["data"][0]["embedding"]

    def get_embedding_and_usage(self, text: str) -> Tuple[List[float], Optional[Dict]]:
        response = self._response(text=text)
        if "data" not in response:
            return [], None

        embedding = response["data"][0]["embedding"]
        usage = response.get("usage", None)
        return embedding, usage



from typing import Optional
from typing_extensions import Literal

from pydantic import Field
from kubernetes.client.models.v1_topology_spread_constraint import (
    V1TopologySpreadConstraint,
)

from phi.k8s.resource.meta.v1.label_selector import LabelSelector
from phi.k8s.resource.base import K8sObject


class TopologySpreadConstraint(K8sObject):
    """
    TopologySpreadConstraint specifies how to spread matching pods among the given topology.

    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#topologyspreadconstraint-v1-core
    """

    resource_type: str = "TopologySpreadConstraint"

    # LabelSelector is used to find matching pods. Pods that match this label selector are counted
    # to determine the number of pods in their corresponding topology domain.
    label_selector: Optional[LabelSelector] = Field(None, alias="labelSelector")
    # MaxSkew describes the degree to which pods may be unevenly distributed.
    # When `whenUnsatisfiable=DoNotSchedule`, it is the maximum permitted difference between the number of matching
    # pods in the target topology and the global minimum.
    # For example, in a 3-zone cluster, MaxSkew is set to 1, and pods with the same labelSelector
    # spread as 1/1/0: | zone1 | zone2 | zone3 | | P | P | | - if MaxSkew is 1, incoming pod can only be scheduled to
    # zone3 to become 1/1/1; scheduling it onto zone1(zone2) would make the ActualSkew(2-0) on zone1(zone2)
    # violate MaxSkew(1). - if MaxSkew is 2, incoming pod can be scheduled onto any zone.
    # When `whenUnsatisfiable=ScheduleAnyway`, it is used to give higher precedence to topologies that satisfy it.
    # It's a required field.
    # Default value is 1 and 0 is not allowed.
    max_skew: Optional[int] = Field(None, alias="maxSkew")
    # TopologyKey is the key of node labels.
    # Nodes that have a label with this key and identical values are considered to be in the same topology.
    # We consider each <key, value> as a "bucket", and try to put balanced number of pods into each bucket.
    # It's a required field.
    topology_key: Optional[str] = Field(None, alias="topologyKey")
    # WhenUnsatisfiable indicates how to deal with a pod if it doesn't satisfy the spread constraint.
    # - DoNotSchedule (default) tells the scheduler not to schedule it.
    # - ScheduleAnyway tells the scheduler to schedule the pod in any location, but giving higher precedence
    # to topologies that would help reduce the skew.
    # A constraint is considered "Unsatisfiable" for an incoming pod if and only if every possible node assignment
    # for that pod would violate "MaxSkew" on some topology.
    # For example, in a 3-zone cluster, MaxSkew is set to 1, and pods with the same labelSelector
    # spread as 3/1/1: | zone1 | zone2 | zone3 | | P P P | P | P |
    # If WhenUnsatisfiable is set to DoNotSchedule, incoming pod can only be scheduled to
    # zone2(zone3) to become 3/2/1(3/1/2) as ActualSkew(2-1) on zone2(zone3) satisfies MaxSkew(1).
    # In other words, the cluster can still be imbalanced, but scheduler won't make it *more* imbalanced.
    # It's a required field. Possible enum values: - `"DoNotSchedule"` instructs the scheduler not to schedule the
    # pod when constraints are not satisfied.
    # - `"ScheduleAnyway"` instructs the scheduler to schedule the pod even if constraints are not satisfied.
    when_unsatisfiable: Optional[Literal["DoNotSchedule", "ScheduleAnyway"]] = Field(None, alias="whenUnsatisfiable")

    def get_k8s_object(self) -> V1TopologySpreadConstraint:
        # Return a V1TopologySpreadConstraint object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_topology_spread_constraint.py
        _v1_topology_spread_constraint = V1TopologySpreadConstraint(
            label_selector=self.label_selector,
            max_skew=self.max_skew,
            topology_key=self.topology_key,
            when_unsatisfiable=self.when_unsatisfiable,
        )
        return _v1_topology_spread_constraint



import json
from typing import Optional, List, Iterator, Dict, Any

from phi.llm.base import LLM
from phi.llm.schemas import Message, FunctionCall
from phi.utils.env import get_from_env
from phi.utils.log import logger
from phi.utils.timer import Timer

try:
    import openai  # noqa: F401
    from openai.openai_object import OpenAIObject  # noqa: F401
except ImportError:
    logger.error("`openai` not installed")
    raise


class OpenAIChat(LLM):
    name: str = "OpenAIChat"
    model: str = "gpt-3.5-turbo-16k"
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    frequency_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None
    stop: Optional[List[str]] = None
    user: Optional[str] = None
    top_p: Optional[float] = None
    logit_bias: Optional[Any] = None
    headers: Optional[Dict[str, Any]] = None

    def to_dict(self) -> Dict[str, Any]:
        _dict = super().to_dict()
        _dict["max_tokens"] = self.max_tokens
        _dict["temperature"] = self.temperature
        return _dict

    @property
    def api_kwargs(self) -> Dict[str, Any]:
        kwargs: Dict[str, Any] = {}
        if self.max_tokens:
            kwargs["max_tokens"] = self.max_tokens
        if self.temperature:
            kwargs["temperature"] = self.temperature
        if self.functions:
            kwargs["functions"] = [f.to_dict() for f in self.functions.values()]
            if self.function_call is not None:
                kwargs["function_call"] = self.function_call
        if self.frequency_penalty:
            kwargs["frequency_penalty"] = self.frequency_penalty
        if self.presence_penalty:
            kwargs["presence_penalty"] = self.presence_penalty
        if self.stop:
            kwargs["stop"] = self.stop
        if self.user:
            kwargs["user"] = self.user
        if self.top_p:
            kwargs["top_p"] = self.top_p
        if self.logit_bias:
            kwargs["logit_bias"] = self.logit_bias
        if self.headers:
            kwargs["headers"] = self.headers
        return kwargs

    def invoke_model(self, messages: List[Message]) -> OpenAIObject:
        if get_from_env("OPENAI_API_KEY") is None:
            logger.debug("--o-o-- Using Phidata Servers")
            try:
                from phi.api.llm import openai_chat

                response_dict = openai_chat(
                    params={
                        "model": self.model,
                        "messages": [m.to_dict() for m in messages],
                        **self.api_kwargs,
                    }
                )
                if response_dict is None:
                    logger.error("Error: Could not reach Phidata Servers.")
                    logger.info("Please message us on https://discord.gg/4MtYHHrgA8 for help.")
                return OpenAIObject.construct_from(response_dict)
            except Exception as e:
                logger.exception(e)
                logger.info("Please message us on https://discord.gg/4MtYHHrgA8 for help.")
                exit(1)
        else:
            return openai.ChatCompletion.create(
                model=self.model,
                messages=[m.to_dict() for m in messages],
                **self.api_kwargs,
            )

    def invoke_model_stream(self, messages: List[Message]) -> Iterator[OpenAIObject]:
        if get_from_env("OPENAI_API_KEY") is None:
            logger.debug("--o-o-- Using Phidata Servers")
            try:
                from phi.api.llm import openai_chat_stream
                from openai import util as openai_util

                for chunk in openai_chat_stream(
                    params={
                        "model": self.model,
                        "messages": [m.to_dict() for m in messages],
                        "stream": True,
                        **self.api_kwargs,
                    }
                ):
                    if chunk:
                        # Sometimes we get multiple chunks in one response which is not valid JSON
                        if "}{" in chunk:
                            # logger.debug(f"Double chunk: {chunk}")
                            chunks = "[" + chunk.replace("}{", "},{") + "]"
                            for chunk_dict in json.loads(chunks, object_hook=openai_util.convert_to_openai_object):
                                yield chunk_dict
                        else:
                            yield json.loads(chunk, object_hook=openai_util.convert_to_openai_object)
            except Exception as e:
                logger.exception(e)
                logger.info("Please message us on https://discord.gg/4MtYHHrgA8 for help.")
                exit(1)
        else:
            yield from openai.ChatCompletion.create(
                model=self.model,
                messages=[m.to_dict() for m in messages],
                stream=True,
                **self.api_kwargs,
            )

    def parsed_response(self, messages: List[Message]) -> str:
        logger.debug("---------- OpenAI Response Start ----------")
        # -*- Log messages for debugging
        for m in messages:
            m.log()

        response_timer = Timer()
        response_timer.start()
        response: OpenAIObject = self.invoke_model(messages=messages)
        response_timer.stop()
        logger.debug(f"Time to generate response: {response_timer.elapsed:.4f}s")
        # logger.debug(f"OpenAI response type: {type(response)}")
        # logger.debug(f"OpenAI response: {response}")

        # -*- Parse response
        response_message = response.choices[0].message
        response_role = response_message.get("role")
        response_content = response_message.get("content")
        response_function_call = response_message.get("function_call")

        # -*- Create assistant message
        assistant_message = Message(
            role=response_role or "assistant",
            content=response_content,
        )
        if response_function_call is not None and isinstance(response_function_call, OpenAIObject):
            assistant_message.function_call = response_function_call.to_dict()

        # -*- Update usage metrics
        # Add response time to metrics
        assistant_message.metrics["time"] = response_timer.elapsed
        if "response_times" not in self.metrics:
            self.metrics["response_times"] = []
        self.metrics["response_times"].append(response_timer.elapsed)

        # Add token usage to metrics
        response_usage = response.usage
        prompt_tokens = response_usage.get("prompt_tokens")
        if prompt_tokens is not None:
            assistant_message.metrics["prompt_tokens"] = prompt_tokens
            if "prompt_tokens" not in self.metrics:
                self.metrics["prompt_tokens"] = prompt_tokens
            else:
                self.metrics["prompt_tokens"] += prompt_tokens
        completion_tokens = response_usage.get("completion_tokens")
        if completion_tokens is not None:
            assistant_message.metrics["completion_tokens"] = completion_tokens
            if "completion_tokens" not in self.metrics:
                self.metrics["completion_tokens"] = completion_tokens
            else:
                self.metrics["completion_tokens"] += completion_tokens
        total_tokens = response_usage.get("total_tokens")
        if total_tokens is not None:
            assistant_message.metrics["total_tokens"] = total_tokens
            if "total_tokens" not in self.metrics:
                self.metrics["total_tokens"] = total_tokens
            else:
                self.metrics["total_tokens"] += total_tokens

        # -*- Add assistant message to messages
        messages.append(assistant_message)
        assistant_message.log()

        # -*- Return content if present, otherwise run function call
        if assistant_message.content is not None:
            return assistant_message.content

        # -*- Parse and run function call
        if assistant_message.function_call is not None and self.run_function_calls:
            _function_name = assistant_message.function_call.get("name")
            _function_arguments_str = assistant_message.function_call.get("arguments")
            if _function_name is not None:
                # Get function call
                function_call = self.get_function_call(name=_function_name, arguments=_function_arguments_str)
                if function_call is None:
                    return "Something went wrong, please try again."

                if self.function_call_stack is None:
                    self.function_call_stack = []

                # -*- Check function call limit
                if len(self.function_call_stack) > self.function_call_limit:
                    return f"Function call limit ({self.function_call_limit}) exceeded."

                # -*- Run function call
                self.function_call_stack.append(function_call)
                function_call_timer = Timer()
                function_call_timer.start()
                function_call.run()
                function_call_timer.stop()
                function_call_message = Message(
                    role="function",
                    name=function_call.function.name,
                    content=function_call.result,
                    metrics={"time": function_call_timer.elapsed},
                )
                messages.append(function_call_message)
                if "function_call_times" not in self.metrics:
                    self.metrics["function_call_times"] = {}
                if function_call.function.name not in self.metrics["function_call_times"]:
                    self.metrics["function_call_times"][function_call.function.name] = []
                self.metrics["function_call_times"][function_call.function.name].append(function_call_timer.elapsed)

                # -*- Get new response using result of function call
                final_response = ""
                if self.show_function_calls:
                    final_response += f"Running: {function_call.get_call_str()}\n\n"
                final_response += self.parsed_response(messages=messages)
                return final_response
        logger.debug("---------- OpenAI Response End ----------")
        return "Something went wrong, please try again."

    def response_message(self, messages: List[Message]) -> Dict:
        logger.debug("---------- OpenAI Response Start ----------")
        # -*- Log messages for debugging
        for m in messages:
            m.log()

        response_timer = Timer()
        response_timer.start()
        response: OpenAIObject = self.invoke_model(messages=messages)
        response_timer.stop()
        logger.debug(f"Time to generate response: {response_timer.elapsed:.4f}s")
        # logger.debug(f"OpenAI response type: {type(response)}")
        # logger.debug(f"OpenAI response: {response}")

        # -*- Parse response
        response_message = response.choices[0].message
        response_role = response_message.get("role")
        response_content = response_message.get("content")
        response_function_call = response_message.get("function_call")

        # -*- Create assistant message
        assistant_message = Message(
            role=response_role or "assistant",
            content=response_content,
        )
        if response_function_call is not None and isinstance(response_function_call, OpenAIObject):
            assistant_message.function_call = response_function_call.to_dict()

        # -*- Update usage metrics
        # Add response time to metrics
        assistant_message.metrics["time"] = response_timer.elapsed
        if "response_times" not in self.metrics:
            self.metrics["response_times"] = []
        self.metrics["response_times"].append(response_timer.elapsed)

        # Add token usage to metrics
        response_usage = response.usage
        prompt_tokens = response_usage.get("prompt_tokens")
        if prompt_tokens is not None:
            assistant_message.metrics["prompt_tokens"] = prompt_tokens
            if "prompt_tokens" not in self.metrics:
                self.metrics["prompt_tokens"] = prompt_tokens
            else:
                self.metrics["prompt_tokens"] += prompt_tokens
        completion_tokens = response_usage.get("completion_tokens")
        if completion_tokens is not None:
            assistant_message.metrics["completion_tokens"] = completion_tokens
            if "completion_tokens" not in self.metrics:
                self.metrics["completion_tokens"] = completion_tokens
            else:
                self.metrics["completion_tokens"] += completion_tokens
        total_tokens = response_usage.get("total_tokens")
        if total_tokens is not None:
            assistant_message.metrics["total_tokens"] = total_tokens
            if "total_tokens" not in self.metrics:
                self.metrics["total_tokens"] = total_tokens
            else:
                self.metrics["total_tokens"] += total_tokens

        # -*- Add assistant message to messages
        messages.append(assistant_message)
        assistant_message.log()

        # -*- Return response
        response_message_dict = response_message.to_dict_recursive()
        logger.debug("---------- OpenAI Response End ----------")
        return response_message_dict

    def parsed_response_stream(self, messages: List[Message]) -> Iterator[str]:
        logger.debug("---------- OpenAI Response Start ----------")
        # -*- Log messages for debugging
        for m in messages:
            m.log()

        assistant_message_content = ""
        assistant_message_function_name = ""
        assistant_message_function_arguments_str = ""
        completion_tokens = 0
        response_timer = Timer()
        response_timer.start()
        for response in self.invoke_model_stream(messages=messages):
            # logger.debug(f"OpenAI response type: {type(response)}")
            # logger.debug(f"OpenAI response: {response}")
            completion_tokens += 1

            # -*- Parse response
            response_delta = response.choices[0].delta
            response_content = response_delta.get("content")
            response_function_call = response_delta.get("function_call")

            # -*- Return content if present, otherwise get function call
            if response_content is not None:
                assistant_message_content += response_content
                yield response_content

            # -*- Parse function call
            if response_function_call is not None:
                _function_name_stream = response_function_call.get("name")
                if _function_name_stream is not None:
                    assistant_message_function_name += _function_name_stream
                _function_args_stream = response_function_call.get("arguments")
                if _function_args_stream is not None:
                    assistant_message_function_arguments_str += _function_args_stream

        response_timer.stop()
        logger.debug(f"Time to generate response: {response_timer.elapsed:.4f}s")

        # -*- Create assistant message
        assistant_message = Message(role="assistant")
        # -*- Add content to assistant message
        if assistant_message_content != "":
            assistant_message.content = assistant_message_content
        # -*- Add function call to assistant message
        if assistant_message_function_name != "":
            assistant_message.function_call = {
                "name": assistant_message_function_name,
                "arguments": assistant_message_function_arguments_str,
            }

        # -*- Update usage metrics
        # Add response time to metrics
        assistant_message.metrics["time"] = response_timer.elapsed
        if "response_times" not in self.metrics:
            self.metrics["response_times"] = []
        self.metrics["response_times"].append(response_timer.elapsed)

        # Add token usage to metrics
        # TODO: compute prompt tokens
        prompt_tokens = 0
        assistant_message.metrics["prompt_tokens"] = prompt_tokens
        if "prompt_tokens" not in self.metrics:
            self.metrics["prompt_tokens"] = prompt_tokens
        else:
            self.metrics["prompt_tokens"] += prompt_tokens
        logger.debug(f"Estimated completion tokens: {completion_tokens}")
        assistant_message.metrics["completion_tokens"] = completion_tokens
        if "completion_tokens" not in self.metrics:
            self.metrics["completion_tokens"] = completion_tokens
        else:
            self.metrics["completion_tokens"] += completion_tokens
        total_tokens = prompt_tokens + completion_tokens
        assistant_message.metrics["total_tokens"] = total_tokens
        if "total_tokens" not in self.metrics:
            self.metrics["total_tokens"] = total_tokens
        else:
            self.metrics["total_tokens"] += total_tokens

        # -*- Add assistant message to messages
        messages.append(assistant_message)
        assistant_message.log()

        # -*- Parse and run function call
        if assistant_message.function_call is not None and self.run_function_calls:
            _function_name = assistant_message.function_call.get("name")
            _function_arguments_str = assistant_message.function_call.get("arguments")
            if _function_name is not None:
                # Get function call
                function_call: Optional[FunctionCall] = self.get_function_call(
                    name=_function_name, arguments=_function_arguments_str
                )
                if function_call is None:
                    return "Something went wrong, please try again."

                if self.function_call_stack is None:
                    self.function_call_stack = []

                # -*- Check function call limit
                if len(self.function_call_stack) > self.function_call_limit:
                    return f"Function call limit ({self.function_call_limit}) exceeded."

                # -*- Run function call
                self.function_call_stack.append(function_call)
                if self.show_function_calls:
                    yield f"Running: {function_call.get_call_str()}\n\n"
                function_call_timer = Timer()
                function_call_timer.start()
                function_call.run()
                function_call_timer.stop()
                function_call_message = Message(
                    role="function",
                    name=function_call.function.name,
                    content=function_call.result,
                    metrics={"time": function_call_timer.elapsed},
                )
                messages.append(function_call_message)
                if "function_call_times" not in self.metrics:
                    self.metrics["function_call_times"] = {}
                if function_call.function.name not in self.metrics["function_call_times"]:
                    self.metrics["function_call_times"][function_call.function.name] = []
                self.metrics["function_call_times"][function_call.function.name].append(function_call_timer.elapsed)

                # -*- Yield new response using result of function call
                yield from self.parsed_response_stream(messages=messages)
        logger.debug("---------- OpenAI Response End ----------")

    def response_delta(self, messages: List[Message]) -> Iterator[Dict]:
        logger.debug("---------- OpenAI Response Start ----------")
        # -*- Log messages for debugging
        for m in messages:
            m.log()

        assistant_message_content = ""
        assistant_message_function_name = ""
        assistant_message_function_arguments_str = ""
        completion_tokens = 0
        response_timer = Timer()
        response_timer.start()
        for response in self.invoke_model_stream(messages=messages):
            # logger.debug(f"OpenAI response type: {type(response)}")
            # logger.debug(f"OpenAI response: {response}")
            completion_tokens += 1

            # -*- Parse response
            response_delta = response.choices[0].delta

            # -*- Read content
            response_content = response_delta.get("content")
            if response_content is not None:
                assistant_message_content += response_content

            # -*- Read function call
            response_function_call = response_delta.get("function_call")
            if response_function_call is not None:
                _function_name_stream = response_function_call.get("name")
                if _function_name_stream is not None:
                    assistant_message_function_name += _function_name_stream
                _function_args_stream = response_function_call.get("arguments")
                if _function_args_stream is not None:
                    assistant_message_function_arguments_str += _function_args_stream

            yield response_delta.to_dict_recursive()
        response_timer.stop()
        logger.debug(f"Time to generate response: {response_timer.elapsed:.4f}s")

        # -*- Create assistant message
        assistant_message = Message(role="assistant")
        # -*- Add content to assistant message
        if assistant_message_content != "":
            assistant_message.content = assistant_message_content
        # -*- Add function call to assistant message
        if assistant_message_function_name != "":
            assistant_message.function_call = {
                "name": assistant_message_function_name,
                "arguments": assistant_message_function_arguments_str,
            }

        # -*- Update usage metrics
        # Add response time to metrics
        assistant_message.metrics["time"] = response_timer.elapsed
        if "response_times" not in self.metrics:
            self.metrics["response_times"] = []
        self.metrics["response_times"].append(response_timer.elapsed)

        # Add token usage to metrics
        # TODO: compute prompt tokens
        prompt_tokens = 0
        assistant_message.metrics["prompt_tokens"] = prompt_tokens
        if "prompt_tokens" not in self.metrics:
            self.metrics["prompt_tokens"] = prompt_tokens
        else:
            self.metrics["prompt_tokens"] += prompt_tokens
        logger.debug(f"Estimated completion tokens: {completion_tokens}")
        assistant_message.metrics["completion_tokens"] = completion_tokens
        if "completion_tokens" not in self.metrics:
            self.metrics["completion_tokens"] = completion_tokens
        else:
            self.metrics["completion_tokens"] += completion_tokens

        total_tokens = prompt_tokens + completion_tokens
        assistant_message.metrics["total_tokens"] = total_tokens
        if "total_tokens" not in self.metrics:
            self.metrics["total_tokens"] = total_tokens
        else:
            self.metrics["total_tokens"] += total_tokens

        # -*- Add assistant message to messages
        messages.append(assistant_message)
        assistant_message.log()
        logger.debug("---------- OpenAI Response End ----------")

    def run_function_call(self, function_call: Dict[str, Any]) -> Optional[Message]:
        _function_name = function_call.get("name")
        _function_arguments_str = function_call.get("arguments")
        if _function_name is not None:
            function_call_obj: Optional[FunctionCall] = self.get_function_call(
                name=_function_name, arguments=_function_arguments_str
            )
            if function_call_obj is None:
                return None

            # -*- Run function call
            function_call_timer = Timer()
            function_call_timer.start()
            function_call_obj.run()
            function_call_timer.stop()
            function_call_message = Message(
                role="function",
                name=function_call_obj.function.name,
                content=function_call_obj.result,
                metrics={"time": function_call_timer.elapsed},
            )
            return function_call_message
        return None



from pathlib import Path
from typing import Optional, List

from typer import launch as typer_launch

from phi.cli.settings import phi_cli_settings, PHI_CLI_DIR
from phi.cli.config import PhiCliConfig
from phi.cli.console import print_info, print_heading
from phi.infra.type import InfraType
from phi.infra.resources import InfraResources
from phi.utils.log import logger


def delete_phidata_conf() -> None:
    from phi.utils.filesystem import delete_from_fs

    logger.debug("Removing existing Phidata configuration")
    delete_from_fs(PHI_CLI_DIR)


def authenticate_user() -> None:
    """Authenticate the user using credentials from phidata.com
    Steps:
    1. Authenticate the user by opening the phidata sign-in url
        and the web-app will post an auth token to a mini http server
        running on the auth_server_port.
    2. Using the auth_token, authenticate the CLI with api and
        save the auth_token. This step is handled by authenticate_and_get_user()
    3. After the user is authenticated update the PhiCliConfig.
    """
    from phi.api.user import authenticate_and_get_user
    from phi.api.schemas.user import UserSchema
    from phi.cli.auth_server import (
        get_port_for_auth_server,
        get_auth_token_from_web_flow,
    )

    print_heading("Authenticating with phidata.com ...")

    auth_server_port = get_port_for_auth_server()
    redirect_uri = "http%3A%2F%2Flocalhost%3A{}%2F".format(auth_server_port)
    auth_url = "{}?source=cli&action=signin&redirecturi={}".format(phi_cli_settings.signin_url, redirect_uri)
    print_info("\nYour browser will be opened to visit:\n{}".format(auth_url))
    typer_launch(auth_url)
    print_info("\nWaiting for a response from browser...\n")

    tmp_auth_token = get_auth_token_from_web_flow(auth_server_port)
    if tmp_auth_token is None:
        logger.error("Could not authenticate, please try again")
        return

    phi_config: Optional[PhiCliConfig] = PhiCliConfig.from_saved_config()
    existing_user: Optional[UserSchema] = phi_config.user if phi_config is not None else None
    try:
        user: Optional[UserSchema] = authenticate_and_get_user(
            tmp_auth_token=tmp_auth_token, existing_user=existing_user
        )
    except Exception as e:
        logger.exception(e)
        logger.error("Could not authenticate, please try again")
        return

    if user is None:
        logger.error("Could not authenticate, please try again")
        return

    if phi_config is None:
        phi_config = PhiCliConfig(user)
        phi_config.save_config()
    else:
        phi_config.user = user

    print_info("Welcome {}".format(user.email))


def initialize_phi(reset: bool = False, login: bool = False) -> bool:
    """Initialize phi on the users machine.

    Steps:
    1. Check if PHI_CLI_DIR exists, if not, create it. If reset == True, recreate PHI_CLI_DIR.
    2. Authenticates the user if login == True.
    3. If PhiCliConfig exists and auth is valid, return True.
    """
    from phi.utils.filesystem import delete_from_fs
    from phi.api.user import create_anon_user

    print_heading("Welcome to phidata!")
    if reset:
        delete_phidata_conf()

    logger.debug("Initializing phidata")

    # Check if ~/.phi exists, if it is not a dir - delete it and create the dir
    if PHI_CLI_DIR.exists():
        logger.debug(f"{PHI_CLI_DIR} exists")
        if not PHI_CLI_DIR.is_dir():
            try:
                delete_from_fs(PHI_CLI_DIR)
            except Exception as e:
                logger.exception(e)
                raise Exception(f"Something went wrong, please delete {PHI_CLI_DIR} and run again")
            PHI_CLI_DIR.mkdir(parents=True, exist_ok=True)
    else:
        PHI_CLI_DIR.mkdir(parents=True)
        logger.debug(f"Created {PHI_CLI_DIR}")

    # Confirm PHI_CLI_DIR exists otherwise we should return
    if PHI_CLI_DIR.exists():
        logger.debug(f"Phidata config location: {PHI_CLI_DIR}")
    else:
        raise Exception("Something went wrong, please try again")

    phi_config: Optional[PhiCliConfig] = PhiCliConfig.from_saved_config()
    if phi_config is None:
        logger.debug("Creating new PhiCliConfig")
        phi_config = PhiCliConfig()
        phi_config.save_config()

    # Authenticate user
    if login:
        authenticate_user()
    else:
        anon_user = create_anon_user()
        if anon_user is not None and phi_config is not None:
            phi_config.user = anon_user

    if phi_config is not None:
        logger.debug("Phidata initialized")
        return True
    else:
        logger.error("Something went wrong, please try again")
        return False


def sign_in_using_cli() -> None:
    from getpass import getpass
    from phi.api.user import sign_in_user
    from phi.api.schemas.user import UserSchema, EmailPasswordAuthSchema

    print_heading("Log in")
    email_raw = input("email: ")
    pass_raw = getpass()

    if email_raw is None or pass_raw is None:
        logger.error("Incorrect email or password")

    try:
        user: Optional[UserSchema] = sign_in_user(EmailPasswordAuthSchema(email=email_raw, password=pass_raw))
    except Exception as e:
        logger.exception(e)
        logger.error("Could not authenticate, please try again")
        return

    if user is None:
        logger.error("Could not get user, please try again")
        return

    phi_config: Optional[PhiCliConfig] = PhiCliConfig.from_saved_config()
    if phi_config is None:
        phi_config = PhiCliConfig(user)
        phi_config.save_config()
    else:
        phi_config.user = user

    print_info("Welcome {}".format(user.email))


def start_resources(
    phi_config: PhiCliConfig,
    resources_file_path: Path,
    target_env: Optional[str] = None,
    target_infra: Optional[InfraType] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
) -> None:
    print_heading(f"Starting resources in: {resources_file_path}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_infra : {target_infra}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    logger.debug(f"\tforce        : {force}")

    from phi.workspace.config import WorkspaceConfig

    if not resources_file_path.exists():
        logger.error(f"File does not exist: {resources_file_path}")
        return

    # Get resource groups to deploy
    resource_groups_to_create: List[InfraResources] = WorkspaceConfig.get_resources_from_file(
        resource_file=resources_file_path,
        env=target_env,
        infra=target_infra,
        order="create",
    )

    # Track number of resource groups created
    num_rgs_created = 0
    num_rgs_to_create = len(resource_groups_to_create)
    # Track number of resources created
    num_resources_created = 0
    num_resources_to_create = 0

    if num_rgs_to_create == 0:
        print_info("No resources to create")
        return

    logger.debug(f"Deploying {num_rgs_to_create} resource groups")
    for rg in resource_groups_to_create:
        _num_resources_created, _num_resources_to_create = rg.create_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
        )
        if _num_resources_created > 0:
            num_rgs_created += 1
        num_resources_created += _num_resources_created
        num_resources_to_create += _num_resources_to_create
        logger.debug(f"Deployed {num_resources_created} resources in {num_rgs_created} resource groups")

    if dry_run:
        return

    if num_resources_created == 0:
        return

    print_heading(f"\n--**-- ResourceGroups deployed: {num_rgs_created}/{num_rgs_to_create}\n")
    if num_resources_created != num_resources_to_create:
        logger.error("Some resources failed to create, please check logs")


def stop_resources(
    phi_config: PhiCliConfig,
    resources_file_path: Path,
    target_env: Optional[str] = None,
    target_infra: Optional[InfraType] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
) -> None:
    print_heading(f"Stopping resources in: {resources_file_path}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_infra : {target_infra}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    logger.debug(f"\tforce        : {force}")

    from phi.workspace.config import WorkspaceConfig

    if not resources_file_path.exists():
        logger.error(f"File does not exist: {resources_file_path}")
        return

    # Get resource groups to shutdown
    resource_groups_to_shutdown: List[InfraResources] = WorkspaceConfig.get_resources_from_file(
        resource_file=resources_file_path,
        env=target_env,
        infra=target_infra,
        order="create",
    )

    # Track number of resource groups deleted
    num_rgs_shutdown = 0
    num_rgs_to_shutdown = len(resource_groups_to_shutdown)
    # Track number of resources created
    num_resources_shutdown = 0
    num_resources_to_shutdown = 0

    if num_rgs_to_shutdown == 0:
        print_info("No resources to delete")
        return

    logger.debug(f"Deleting {num_rgs_to_shutdown} resource groups")
    for rg in resource_groups_to_shutdown:
        _num_resources_shutdown, _num_resources_to_shutdown = rg.delete_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
        )
        if _num_resources_shutdown > 0:
            num_rgs_shutdown += 1
        num_resources_shutdown += _num_resources_shutdown
        num_resources_to_shutdown += _num_resources_to_shutdown
        logger.debug(f"Deleted {num_resources_shutdown} resources in {num_rgs_shutdown} resource groups")

    if dry_run:
        return

    if num_resources_shutdown == 0:
        return

    print_heading(f"\n--**-- ResourceGroups deleted: {num_rgs_shutdown}/{num_rgs_to_shutdown}\n")
    if num_resources_shutdown != num_resources_to_shutdown:
        logger.error("Some resources failed to delete, please check logs")


def patch_resources(
    phi_config: PhiCliConfig,
    resources_file_path: Path,
    target_env: Optional[str] = None,
    target_infra: Optional[InfraType] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
) -> None:
    print_heading(f"Updating resources in: {resources_file_path}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_infra : {target_infra}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    logger.debug(f"\tforce        : {force}")

    from phi.workspace.config import WorkspaceConfig

    if not resources_file_path.exists():
        logger.error(f"File does not exist: {resources_file_path}")
        return

    # Get resource groups to update
    resource_groups_to_patch: List[InfraResources] = WorkspaceConfig.get_resources_from_file(
        resource_file=resources_file_path,
        env=target_env,
        infra=target_infra,
        order="create",
    )

    num_rgs_patched = 0
    num_rgs_to_patch = len(resource_groups_to_patch)
    # Track number of resources updated
    num_resources_patched = 0
    num_resources_to_patch = 0

    if num_rgs_to_patch == 0:
        print_info("No resources to patch")
        return

    logger.debug(f"Deploying {num_rgs_to_patch} resource groups")
    for rg in resource_groups_to_patch:
        _num_resources_patched, _num_resources_to_patch = rg.update_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
        )
        if _num_resources_patched > 0:
            num_rgs_patched += 1
        num_resources_patched += _num_resources_patched
        num_resources_to_patch += _num_resources_to_patch
        logger.debug(f"Deployed {num_resources_patched} resources in {num_rgs_patched} resource groups")

    if dry_run:
        return

    if num_resources_patched == 0:
        return

    print_heading(f"\n--**-- ResourceGroups patched: {num_rgs_patched}/{num_rgs_to_patch}\n")
    if num_resources_patched != num_resources_to_patch:
        logger.error("Some resources failed to patch, please check logs")



import json
from typing import List, Iterator, Optional, Dict, Any, Callable

from pydantic import BaseModel, ConfigDict

from phi.llm.schemas import Message, Function, FunctionCall
from phi.llm.function.registry import FunctionRegistry
from phi.utils.log import logger


class LLM(BaseModel):
    # ID of the model to use.
    model: str
    # Name for this LLM (not used for api calls).
    name: Optional[str] = None
    # Metrics collected for this LLM (not used for api calls).
    metrics: Dict[str, Any] = {}

    # A list of functions the model may generate JSON inputs for.
    # Do not add functions manually. Use add_function() or add_function_schema() instead.
    functions: Optional[Dict[str, Function]] = None
    # Controls how the model calls functions.
    # "none" means the model will not call a function and instead generates a message.
    # "auto" means the model can pick between generating a message or calling a function.
    # Specifying a particular function via {"name": "my_function"} forces the model to call that function.
    # "none" is the default when no functions are present. "auto" is the default if functions are present.
    function_call: Optional[str] = None
    function_call_limit: int = 50
    function_call_stack: Optional[List[FunctionCall]] = None
    show_function_calls: Optional[bool] = None
    # If True, runs function calls before sending back the response content.
    run_function_calls: bool = True

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def api_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError

    def invoke_model(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def invoke_model_stream(self, *args, **kwargs) -> Iterator[Any]:
        raise NotImplementedError

    def parsed_response(self, messages: List[Message]) -> str:
        raise NotImplementedError

    def response_message(self, messages: List[Message]) -> Dict:
        raise NotImplementedError

    def parsed_response_stream(self, messages: List[Message]) -> Iterator[str]:
        raise NotImplementedError

    def response_delta(self, messages: List[Message]) -> Iterator[Dict]:
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        _dict = self.model_dump(include={"model", "name", "metrics"})
        if self.functions:
            _dict["functions"] = {k: v.to_dict() for k, v in self.functions.items()}
            _dict["function_call_limit"] = self.function_call_limit
        return _dict

    def add_function(self, f: Callable) -> None:
        func = Function.from_callable(f)
        if self.functions is None:
            self.functions = {}
        self.functions[func.name] = func
        logger.debug(f"Added function {func.name} to LLM.")

    def add_function_schema(self, func: Function, if_not_exists: bool = True) -> None:
        if self.functions is None:
            self.functions = {}

        if if_not_exists and func.name in self.functions:
            return

        self.functions[func.name] = func
        logger.debug(f"Added function {func.name} to LLM.")

    def add_function_registry(self, registry: FunctionRegistry) -> None:
        if self.functions is None:
            self.functions = {}

        self.functions.update(registry.functions)
        logger.debug(f"Functions from {registry.name} added to LLM.")

    def get_function_call(self, name: str, arguments: Optional[str] = None) -> Optional[FunctionCall]:
        logger.debug(f"Getting function {name}. Args: {arguments}")
        if self.functions is None:
            return None

        function_to_call: Optional[Function] = None
        if name in self.functions:
            function_to_call = self.functions[name]
        if function_to_call is None:
            logger.error(f"Function {name} not found")
            return None

        function_call = FunctionCall(function=function_to_call)
        if arguments is not None and arguments != "":
            try:
                if "None" in arguments:
                    arguments = arguments.replace("None", "null")
                if "True" in arguments:
                    arguments = arguments.replace("True", "true")
                if "False" in arguments:
                    arguments = arguments.replace("False", "false")
                _arguments = json.loads(arguments)
            except Exception as e:
                logger.error(f"Unable to decode function arguments {arguments}: {e}")
                return None

            if not isinstance(_arguments, dict):
                logger.error(f"Function arguments {arguments} is not a valid JSON object")
                return None

            try:
                clean_arguments: Dict[str, Any] = {}
                for k, v in _arguments.items():
                    if isinstance(v, str):
                        _v = v.strip().lower()
                        if _v in ("none", "null"):
                            clean_arguments[k] = None
                        elif _v == "true":
                            clean_arguments[k] = True
                        elif _v == "false":
                            clean_arguments[k] = False
                        else:
                            clean_arguments[k] = v.strip()
                    else:
                        clean_arguments[k] = v

                function_call.arguments = clean_arguments
            except Exception as e:
                logger.error(f"Unable to parse function arguments {arguments}: {e}")
                return None

        return function_call



PYTHONPATH_ENV_VAR: str = "PYTHONPATH"
PHI_RUNTIME_ENV_VAR: str = "PHI_RUNTIME"

SCRIPTS_DIR_ENV_VAR: str = "PHI_SCRIPTS_DIR"
STORAGE_DIR_ENV_VAR: str = "PHI_STORAGE_DIR"
WORKFLOWS_DIR_ENV_VAR: str = "PHI_WORKFLOWS_DIR"
WORKSPACE_NAME_ENV_VAR: str = "PHI_WORKSPACE_NAME"
WORKSPACE_ROOT_ENV_VAR: str = "PHI_WORKSPACE_ROOT"
WORKSPACES_MOUNT_ENV_VAR: str = "PHI_WORKSPACES_MOUNT"
WORKSPACE_ID_ENV_VAR: str = "PHI_WORKSPACE_ID"
WORKSPACE_HASH_ENV_VAR: str = "PHI_WORKSPACE_HASH"
WORKSPACE_DIR_ENV_VAR: str = "PHI_WORKSPACE_DIR"
REQUIREMENTS_FILE_PATH_ENV_VAR: str = "REQUIREMENTS_FILE_PATH"

AWS_REGION_ENV_VAR: str = "AWS_REGION"
AWS_DEFAULT_REGION_ENV_VAR: str = "AWS_DEFAULT_REGION"
AWS_PROFILE_ENV_VAR: str = "AWS_PROFILE"
AWS_CONFIG_FILE_ENV_VAR: str = "AWS_CONFIG_FILE"
AWS_SHARED_CREDENTIALS_FILE_ENV_VAR: str = "AWS_SHARED_CREDENTIALS_FILE"

INIT_AIRFLOW_ENV_VAR: str = "INIT_AIRFLOW"
AIRFLOW_ENV_ENV_VAR: str = "AIRFLOW_ENV"
AIRFLOW_HOME_ENV_VAR: str = "AIRFLOW_HOME"
AIRFLOW_EXECUTOR_ENV_VAR: str = "AIRFLOW__CORE__EXECUTOR"
AIRFLOW_DAGS_FOLDER_ENV_VAR: str = "AIRFLOW__CORE__DAGS_FOLDER"
AIRFLOW_DB_CONN_URL_ENV_VAR: str = "AIRFLOW__DATABASE__SQL_ALCHEMY_CONN"



from collections import OrderedDict
from pathlib import Path
from typing import Dict, List, Optional

from phi.cli.console import print_heading, print_info
from phi.cli.settings import phi_cli_settings
from phi.api.schemas.user import UserSchema
from phi.api.schemas.workspace import WorkspaceSchema, WorkspaceDelete
from phi.utils.log import logger
from phi.utils.json_io import read_json_file, write_json_file
from phi.workspace.config import WorkspaceConfig


class PhiCliConfig:
    """The PhiCliConfig class manages user data for the phi cli"""

    def __init__(
        self,
        user: Optional[UserSchema] = None,
        active_ws_dir: Optional[str] = None,
        ws_config_map: Optional[Dict[str, WorkspaceConfig]] = None,
    ) -> None:
        # Current user, populated after authenticating with the api
        # To add a user, use the user setter
        self._user: Optional[UserSchema] = user

        # Active ws dir - used as the default for `phi` commands
        # To add an active workspace, use the active_ws_dir setter
        self._active_ws_dir: Optional[str] = active_ws_dir

        # Mapping from ws_root_path to ws_config
        self.ws_config_map: Dict[str, WorkspaceConfig] = ws_config_map or OrderedDict()

    ######################################################
    ## User functions
    ######################################################

    @property
    def user(self) -> Optional[UserSchema]:
        return self._user

    @user.setter
    def user(self, user: Optional[UserSchema]) -> None:
        """Sets the user"""
        if user is not None:
            logger.debug(f"Setting user to: {user.email}")
            clear_user_cache = (
                self._user is not None  # previous user is not None
                and self._user.email != "anon"  # previous user is not anon
                and (user.email != self._user.email or user.id_user != self._user.id_user)  # new user is different
            )
            self._user = user
            if clear_user_cache:
                self.clear_user_cache()
            self.save_config()

    def clear_user_cache(self) -> None:
        """Clears the user cache"""
        logger.debug("Clearing user cache")
        self.ws_config_map.clear()
        self._active_ws_dir = None
        phi_cli_settings.ai_conversations_path.unlink(missing_ok=True)
        logger.info("Workspaces cleared, please setup again using `phi ws setup`")

    ######################################################
    ## Workspace functions
    ######################################################

    @property
    def active_ws_dir(self) -> Optional[str]:
        return self._active_ws_dir

    def set_active_ws_dir(self, ws_root_path: Optional[Path]) -> None:
        if ws_root_path is not None:
            logger.debug(f"Setting active workspace to: {str(ws_root_path)}")
            self._active_ws_dir = str(ws_root_path)
            self.save_config()

    @property
    def available_ws(self) -> List[WorkspaceConfig]:
        return list(self.ws_config_map.values())

    def _add_or_update_ws_config(
        self, ws_root_path: Path, ws_schema: Optional[WorkspaceSchema] = None
    ) -> Optional[WorkspaceConfig]:
        """The main function to create, update or refresh a WorkspaceConfig.

        This function does not call self.save_config(). Remember to save_config() after calling this function.
        """

        # Validate ws_root_path
        if ws_root_path is None or not isinstance(ws_root_path, Path):
            raise ValueError(f"Invalid ws_root: {ws_root_path}")
        ws_root_str = str(ws_root_path)

        ######################################################
        # Create new ws_config if one does not exist
        ######################################################
        if ws_root_str not in self.ws_config_map:
            logger.debug(f"Creating workspace at: {ws_root_str}")
            new_workspace_config = WorkspaceConfig(
                ws_root_path=ws_root_path,
                ws_schema=ws_schema,
            )
            self.ws_config_map[ws_root_str] = new_workspace_config
            logger.debug(f"Workspace created at: {ws_root_str}")

            # Return the new_workspace_config
            return new_workspace_config

        ######################################################
        # Update ws_config
        ######################################################
        logger.debug(f"Updating workspace at: {ws_root_str}")
        # By this point there should be a WorkspaceConfig object for this ws_name
        existing_ws_config: Optional[WorkspaceConfig] = self.ws_config_map.get(ws_root_str, None)
        if existing_ws_config is None:
            logger.error(f"Could not find workspace at: {ws_root_str}, please run `phi ws setup`")
            return None

        # Update the ws_schema if it's not None and different from the existing one
        if ws_schema is not None and existing_ws_config.ws_schema != ws_schema:
            existing_ws_config.ws_schema = ws_schema
        logger.debug(f"Workspace updated: {ws_root_str}")

        # Return the updated_ws_config
        return existing_ws_config

    ######################################################
    # END
    ######################################################

    def add_new_ws_to_config(self, ws_root_path: Path) -> Optional[WorkspaceConfig]:
        """Adds a newly created workspace to the PhiCliConfig"""

        ws_config = self._add_or_update_ws_config(ws_root_path=ws_root_path)
        self.save_config()
        return ws_config

    def update_ws_config(
        self,
        ws_root_path: Path,
        ws_schema: Optional[WorkspaceSchema] = None,
        set_as_active: bool = False,
    ) -> Optional[WorkspaceConfig]:
        """Updates WorkspaceConfig and returns True if successful"""
        ws_config = self._add_or_update_ws_config(
            ws_root_path=ws_root_path,
            ws_schema=ws_schema,
        )
        if set_as_active:
            self._active_ws_dir = str(ws_root_path)
        self.save_config()
        return ws_config

    def delete_ws(self, ws_root_path: Path) -> None:
        """Handles Deleting a workspace from the PhiCliConfig and api"""

        ws_root_str = str(ws_root_path)
        print_heading(f"Deleting record for workspace at: {ws_root_str}")
        print_info("-*- Note: this does not delete any files on disk, please delete them manually")

        ws_config: Optional[WorkspaceConfig] = self.ws_config_map.pop(ws_root_str, None)
        if ws_config is None:
            logger.warning(f"No record of workspace at {ws_root_str}")
            return

        # Check if we're deleting the active workspace, if yes, unset the active ws
        if self._active_ws_dir is not None and self._active_ws_dir == ws_root_str:
            print_info(f"Removing {ws_root_str} as the active workspace")
            self._active_ws_dir = None

        if self.user is not None and ws_config.ws_schema is not None:
            print_info("Deleting workspace from the server")

            from phi.api.workspace import delete_workspace_for_user

            delete_workspace_for_user(
                user=self.user,
                workspace=WorkspaceDelete(
                    id_workspace=ws_config.ws_schema.id_workspace, ws_name=ws_config.ws_schema.ws_name
                ),
            )
        self.save_config()

    ######################################################
    ## Get Workspace Data
    ######################################################

    def get_ws_config_by_dir_name(self, ws_dir_name: str) -> Optional[WorkspaceConfig]:
        ws_root_str: Optional[str] = None
        for k, v in self.ws_config_map.items():
            if v.ws_root_path.stem == ws_dir_name:
                ws_root_str = k
                break

        if ws_root_str is None or ws_root_str not in self.ws_config_map:
            return None

        return self.ws_config_map[ws_root_str]

    def get_ws_config_by_path(self, ws_root_path: Path) -> Optional[WorkspaceConfig]:
        return self.ws_config_map[str(ws_root_path)] if str(ws_root_path) in self.ws_config_map else None

    def get_active_ws_config(self) -> Optional[WorkspaceConfig]:
        if self.active_ws_dir is not None and self.active_ws_dir in self.ws_config_map:
            return self.ws_config_map[self.active_ws_dir]
        return None

    ######################################################
    ## Save PhiCliConfig
    ######################################################

    def save_config(self):
        config_data = {
            "user": self.user.model_dump() if self.user else None,
            "active_ws_dir": self.active_ws_dir,
            "ws_config_map": {k: v.to_dict() for k, v in self.ws_config_map.items()},
        }
        write_json_file(file_path=phi_cli_settings.config_file_path, data=config_data)

    @classmethod
    def from_saved_config(cls):
        try:
            config_data = read_json_file(file_path=phi_cli_settings.config_file_path)
            if config_data is None or not isinstance(config_data, dict):
                logger.debug("No config found")
                return cls()

            user_dict = config_data.get("user")
            user_schema = UserSchema.model_validate(user_dict) if user_dict else None
            active_ws_dir = config_data.get("active_ws_dir")

            # Create a new config
            new_config = cls(user_schema, active_ws_dir)

            # Add all the workspaces
            for k, v in config_data.get("ws_config_map", {}).items():
                _ws_config = WorkspaceConfig.from_dict(v)
                if _ws_config is not None:
                    new_config.ws_config_map[k] = _ws_config
            return new_config
        except Exception as e:
            logger.warning(e)
            logger.warning("Please setup the workspace using `phi ws setup`")

    ######################################################
    ## Print PhiCliConfig
    ######################################################

    def print_to_cli(self, show_all: bool = False):
        if self.user:
            print_heading(f"User: {self.user.email}\n")
        if self.active_ws_dir:
            print_heading(f"Active workspace directory: {self.active_ws_dir}\n")
        else:
            print_info("No active workspace found.")
            print_info(
                "Please create a workspace using `phi ws create` " "or setup existing workspace using `phi ws setup`"
            )

        if show_all and len(self.ws_config_map) > 0:
            print_heading("Available workspaces:\n")
            c = 1
            for k, v in self.ws_config_map.items():
                print_info(f"  {c}. {k}")
                if v.ws_schema and v.ws_schema.ws_name:
                    print_info(f"     Name: {v.ws_schema.ws_name}")
                c += 1



from pathlib import Path
from typing import Optional, Dict, List


from phi.api.workspace import log_workspace_event
from phi.api.schemas.workspace import (
    WorkspaceSchema,
    WorkspaceCreate,
    WorkspaceUpdate,
    WorkspaceEvent,
    UpdatePrimaryWorkspace,
)
from phi.cli.config import PhiCliConfig
from phi.cli.console import (
    console,
    print_heading,
    print_info,
    print_subheading,
    log_config_not_available_msg,
)
from phi.infra.type import InfraType
from phi.infra.resources import InfraResources
from phi.workspace.config import WorkspaceConfig
from phi.workspace.enums import WorkspaceStarterTemplate
from phi.utils.log import logger

TEMPLATE_TO_NAME_MAP: Dict[WorkspaceStarterTemplate, str] = {
    WorkspaceStarterTemplate.llm_app: "llm-app",
    WorkspaceStarterTemplate.api_app: "api-app",
    WorkspaceStarterTemplate.django_app: "django-app",
    WorkspaceStarterTemplate.streamlit_app: "streamlit-app",
    WorkspaceStarterTemplate.ai_platform: "ai-platform",
}
TEMPLATE_TO_REPO_MAP: Dict[WorkspaceStarterTemplate, str] = {
    WorkspaceStarterTemplate.llm_app: "https://github.com/phidatahq/llm-app.git",
    WorkspaceStarterTemplate.api_app: "https://github.com/phidatahq/api-app.git",
    WorkspaceStarterTemplate.django_app: "https://github.com/phidatahq/django-app.git",
    WorkspaceStarterTemplate.streamlit_app: "https://github.com/phidatahq/streamlit-app.git",
    WorkspaceStarterTemplate.ai_platform: "https://github.com/phidatahq/ai-platform.git",
}


def create_workspace(name: Optional[str] = None, template: Optional[str] = None, url: Optional[str] = None) -> bool:
    """Creates a new workspace.

    This function clones a template or url on the users machine at the path:
        cwd/name
    """
    import git
    from shutil import copytree
    from rich.prompt import Prompt

    from phi.cli.operator import initialize_phi
    from phi.utils.common import str_to_int
    from phi.utils.filesystem import rmdir_recursive
    from phi.workspace.helpers import get_workspace_dir_path
    from phi.utils.git import GitCloneProgress

    current_dir: Path = Path(".").resolve()

    # Phi should be initialized before creating a workspace
    phi_config: Optional[PhiCliConfig] = PhiCliConfig.from_saved_config()
    if not phi_config:
        init_success = initialize_phi()
        if not init_success:
            from phi.cli.console import log_phi_init_failed_msg

            log_phi_init_failed_msg()
            return False
        phi_config = PhiCliConfig.from_saved_config()
        # If phi_config is still None, throw an error
        if not phi_config:
            log_config_not_available_msg()
            return False

    ws_dir_name: Optional[str] = name
    repo_to_clone: Optional[str] = url
    ws_template = WorkspaceStarterTemplate.llm_app
    templates = list(WorkspaceStarterTemplate.__members__.values())

    if repo_to_clone is None:
        # Get repo_to_clone from template
        if template is None:
            # Get starter template from the user if template is not provided
            # Display available starter templates and ask user to select one
            print_info("Select starter template or press Enter for default (llm-app)")
            for template_id, template_name in enumerate(templates, start=1):
                print_info("  [{}] {}".format(template_id, template_name))

            # Get starter template from the user
            template_choices = [str(idx) for idx, _ in enumerate(templates, start=1)]
            template_inp_raw = Prompt.ask("Template Number", choices=template_choices, default="1", show_choices=False)
            # Convert input to int
            template_inp = str_to_int(template_inp_raw)

            if template_inp is not None:
                template_inp_idx = template_inp - 1
                ws_template = WorkspaceStarterTemplate(templates[template_inp_idx])
        elif template.lower() in WorkspaceStarterTemplate.__members__.values():
            ws_template = WorkspaceStarterTemplate(template)
        else:
            raise Exception(f"{template} is not a supported template, please choose from: {templates}")

        logger.debug(f"Selected Template: {ws_template.value}")
        repo_to_clone = TEMPLATE_TO_REPO_MAP.get(ws_template)

    if ws_dir_name is None:
        default_ws_name = "llm-app"
        if url is not None:
            # Get default_ws_name from url
            default_ws_name = url.split("/")[-1].split(".")[0]
        else:
            # Get default_ws_name from template
            default_ws_name = TEMPLATE_TO_NAME_MAP.get(ws_template, "llm-app")
        logger.debug(f"asking for ws name with default: {default_ws_name}")
        # Ask user for workspace name if not provided
        ws_dir_name = Prompt.ask("Workspace Name", default=default_ws_name, console=console)

    if ws_dir_name is None:
        logger.error("Workspace name is required")
        return False
    if repo_to_clone is None:
        logger.error("URL or Template is required")
        return False

    # Check if we can create the workspace in the current dir
    ws_root_path: Path = current_dir.joinpath(ws_dir_name)
    if ws_root_path.exists():
        logger.error(f"Directory {ws_root_path} exists, please delete directory or choose another name for workspace")
        return False

    print_info(f"Creating {str(ws_root_path)}")
    logger.debug("Cloning: {}".format(repo_to_clone))
    try:
        _cloned_git_repo: git.Repo = git.Repo.clone_from(
            repo_to_clone, str(ws_root_path), progress=GitCloneProgress()  # type: ignore
        )
    except Exception as e:
        logger.error(e)
        return False

    # Remove existing .git folder
    _dot_git_folder = ws_root_path.joinpath(".git")
    _dot_git_exists = _dot_git_folder.exists()
    if _dot_git_exists:
        logger.debug(f"Deleting {_dot_git_folder}")
        try:
            _dot_git_exists = not rmdir_recursive(_dot_git_folder)
        except Exception as e:
            logger.warning(f"Failed to delete {_dot_git_folder}: {e}")
            logger.info("Please delete the .git folder manually")
            pass

    phi_config.add_new_ws_to_config(ws_root_path=ws_root_path)

    try:
        # workspace_dir_path is the path to the ws_root/workspace dir
        workspace_dir_path: Path = get_workspace_dir_path(ws_root_path)
        workspace_secrets_dir = workspace_dir_path.joinpath("secrets").resolve()
        workspace_example_secrets_dir = workspace_dir_path.joinpath("example_secrets").resolve()

        print_info(f"Creating {str(workspace_secrets_dir)}")
        copytree(
            str(workspace_example_secrets_dir),
            str(workspace_secrets_dir),
        )
    except Exception as e:
        logger.warning(f"Could not create workspace/secrets: {e}")
        logger.warning("Please manually copy workspace/example_secrets to workspace/secrets")

    print_info(f"Your new workspace is available at {str(ws_root_path)}\n")
    return setup_workspace(ws_root_path=ws_root_path)


def setup_workspace(ws_root_path: Path) -> bool:
    """Setup a phi workspace at `ws_root_path`.

    1. Validate pre-requisites
    1.1 Check ws_root_path is available
    1.2 Check PhiCliConfig is available
    1.3 Validate WorkspaceConfig is available
    1.4 Load workspace and set as active
    1.5 Check if remote origin is available
    1.6 Create anon user if not available

    2. Create or Update WorkspaceSchema
    If a ws_schema exists for this workspace, this workspace has a record in the backend
    2.1 Create WorkspaceSchema for a NEWLY CREATED WORKSPACE
    2.2 Set workspace as primary if needed
    2.3 Update WorkspaceSchema if git_url has changed
    """
    from phi.cli.operator import initialize_phi
    from phi.utils.git import get_remote_origin_for_dir

    print_heading("Running workspace setup\n")

    ######################################################
    ## 1. Validate Pre-requisites
    ######################################################
    ######################################################
    # 1.1 Check ws_root_path is available
    ######################################################
    _ws_is_valid: bool = ws_root_path is not None and ws_root_path.exists() and ws_root_path.is_dir()
    if not _ws_is_valid:
        logger.error("Invalid directory: {}".format(ws_root_path))
        return False

    ######################################################
    # 1.2 Check PhiCliConfig is available
    ######################################################
    phi_config: Optional[PhiCliConfig] = PhiCliConfig.from_saved_config()
    if not phi_config:
        # Phidata should be initialized before workspace setup
        init_success = initialize_phi()
        if not init_success:
            from phi.cli.console import log_phi_init_failed_msg

            log_phi_init_failed_msg()
            return False
        phi_config = PhiCliConfig.from_saved_config()
        # If phi_config is still None, throw an error
        if not phi_config:
            raise Exception("Failed to initialize phi")

    ######################################################
    # 1.3 Validate WorkspaceConfig is available
    ######################################################
    logger.debug(f"Checking for a workspace at {ws_root_path}")
    ws_config: Optional[WorkspaceConfig] = phi_config.get_ws_config_by_path(ws_root_path)
    if ws_config is None:
        # This happens if
        # - The user is setting up a workspace not previously setup on this machine
        # - OR the user ran `phi init -r` which erases existing records of workspaces
        logger.debug(f"Could not find an existing workspace at: {ws_root_path}")

        # In this case, the local workspace directory exists but PhiCliConfig does not have a record
        print_info(f"Adding {str(ws_root_path.stem)} as a workspace")
        phi_config.add_new_ws_to_config(ws_root_path=ws_root_path)
        ws_config = phi_config.get_ws_config_by_path(ws_root_path)
    else:
        logger.debug(f"Found workspace at {ws_root_path}")

    # If the ws_config is still None it means the workspace is corrupt
    if ws_config is None:
        logger.error(f"Could not find workspace at: {str(ws_root_path)}")
        logger.error("Please try again")
        return False

    ######################################################
    # 1.4 Load workspace and set as active
    ######################################################
    # Load and save the workspace config
    # ws_config.load()
    # Get the workspace dir name
    ws_dir_name = ws_config.ws_root_path.stem
    # Set the workspace as active if it is not already
    # update_primary_ws is a flag to update the primary workspace in the backend
    update_primary_ws = False
    if phi_config.active_ws_dir is None or phi_config.active_ws_dir != ws_dir_name:
        phi_config.set_active_ws_dir(ws_config.ws_root_path)
        update_primary_ws = True

    ######################################################
    # 1.5 Check if remote origin is available
    ######################################################
    git_remote_origin_url: Optional[str] = get_remote_origin_for_dir(ws_root_path)
    logger.debug("Git origin: {}".format(git_remote_origin_url))

    ######################################################
    # 1.6 Create anon user if not logged in
    ######################################################
    if phi_config.user is None:
        from phi.api.user import create_anon_user

        logger.debug("Creating anon user")
        anon_user = create_anon_user()
        if anon_user is not None:
            phi_config.user = anon_user

    ######################################################
    ## 2. Create or Update WorkspaceSchema
    ######################################################
    # If a ws_schema exists for this workspace, this workspace is synced with the api
    ws_schema: Optional[WorkspaceSchema] = ws_config.ws_schema
    if phi_config.user is not None:
        ######################################################
        # 2.1 Create WorkspaceSchema for NEW WORKSPACE
        ######################################################
        if ws_schema is None or ws_schema.id_workspace is None:
            from phi.api.workspace import create_workspace_for_user
            from phi.workspace.helpers import generate_workspace_name

            # If ws_schema is None, this is a NEWLY CREATED WORKSPACE.
            # We make a call to the api to create a new ws_schema
            new_workspace_name = generate_workspace_name(ws_dir_name=ws_dir_name)
            logger.debug("Creating ws_schema for new workspace")
            logger.debug(f"ws_dir_name: {ws_dir_name}")
            logger.debug(f"workspace_name: {new_workspace_name}")

            ws_schema = create_workspace_for_user(
                user=phi_config.user,
                workspace=WorkspaceCreate(
                    ws_name=new_workspace_name,
                    git_url=git_remote_origin_url,
                    is_primary_for_user=True,
                ),
            )
            if ws_schema is not None:
                ws_config = phi_config.update_ws_config(ws_root_path=ws_root_path, ws_schema=ws_schema)
            else:
                logger.debug("Failed to sync workspace with api. Please setup again")

        ######################################################
        # 2.2 Set workspace as primary if needed
        ######################################################
        elif update_primary_ws:
            from phi.api.workspace import update_primary_workspace_for_user

            logger.debug("Setting workspace as primary")
            logger.debug(f"ws_dir_name: {ws_dir_name}")
            logger.debug(f"workspace_name: {ws_schema.ws_name}")

            updated_workspace_schema = update_primary_workspace_for_user(
                user=phi_config.user,
                workspace=UpdatePrimaryWorkspace(
                    id_workspace=ws_schema.id_workspace,
                    ws_name=ws_schema.ws_name,
                ),
            )

            if updated_workspace_schema is not None:
                # Update the ws_schema for this workspace.
                ws_config = phi_config.update_ws_config(ws_root_path=ws_root_path, ws_schema=updated_workspace_schema)
            else:
                logger.debug("Failed to sync workspace with api. Please setup again")

        ######################################################
        # 2.3 Update WorkspaceSchema if git_url has changed
        ######################################################
        if ws_schema is not None and ws_schema.git_url != git_remote_origin_url:
            from phi.api.workspace import update_workspace_for_user

            logger.debug("Updating git_url for existing workspace")
            logger.debug(f"ws_dir_name: {ws_dir_name}")
            logger.debug(f"workspace_name: {ws_schema.ws_name}")
            logger.debug(f"Existing git_url: {ws_schema.git_url}")
            logger.debug(f"New git_url: {git_remote_origin_url}")

            updated_workspace_schema = update_workspace_for_user(
                user=phi_config.user,
                workspace=WorkspaceUpdate(
                    id_workspace=ws_schema.id_workspace,
                    git_url=git_remote_origin_url,
                ),
            )
            if updated_workspace_schema is not None:
                # Update the ws_schema for this workspace.
                ws_config = phi_config.update_ws_config(ws_root_path=ws_root_path, ws_schema=updated_workspace_schema)
            else:
                logger.debug("Failed to sync workspace with api. Please setup again")

    if ws_config is not None:
        # logger.debug("Workspace Config: {}".format(ws_config.model_dump_json(indent=2)))
        print_subheading("Setup complete! Next steps:")
        print_info("1. Start workspace:")
        print_info("\tphi ws up")
        print_info("2. Stop workspace:")
        print_info("\tphi ws down")
        if ws_config.workspace_settings is not None:
            scripts_dir = ws_config.workspace_settings.scripts_dir
            install_ws_file = f"sh {ws_root_path}/{scripts_dir}/install.sh"
            print_info("3. Install workspace dependencies:")
            print_info(f"\t{install_ws_file}")

        if ws_config.ws_schema is not None and phi_config.user is not None:
            log_workspace_event(
                user=phi_config.user,
                workspace_event=WorkspaceEvent(
                    id_workspace=ws_config.ws_schema.id_workspace,
                    event_type="setup",
                    event_status="success",
                    event_data={"workspace_root_path": str(ws_root_path)},
                ),
            )
        return True
    else:
        print_info("Workspace setup unsuccessful. Please try again.")
    return False
    ######################################################
    ## End Workspace setup
    ######################################################


def start_workspace(
    phi_config: PhiCliConfig,
    ws_config: WorkspaceConfig,
    target_env: Optional[str] = None,
    target_infra: Optional[InfraType] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
) -> None:
    """Start a Phi Workspace. This is called from `phi ws up`"""
    if ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return

    # Set the local environment variables before processing configs
    ws_config.set_local_env()

    # Get resource groups to deploy
    resource_groups_to_create: List[InfraResources] = ws_config.get_resources(
        env=target_env,
        infra=target_infra,
        order="create",
    )

    # Track number of resource groups created
    num_rgs_created = 0
    num_rgs_to_create = len(resource_groups_to_create)
    # Track number of resources created
    num_resources_created = 0
    num_resources_to_create = 0

    if num_rgs_to_create == 0:
        print_info("No resources to create")
        return

    logger.debug(f"Deploying {num_rgs_to_create} resource groups")
    for rg in resource_groups_to_create:
        _num_resources_created, _num_resources_to_create = rg.create_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
        )
        if _num_resources_created > 0:
            num_rgs_created += 1
        num_resources_created += _num_resources_created
        num_resources_to_create += _num_resources_to_create
        logger.debug(f"Deployed {num_resources_created} resources in {num_rgs_created} resource groups")

    if dry_run:
        return

    if num_resources_created == 0:
        return

    print_heading(f"\n--**-- ResourceGroups deployed: {num_rgs_created}/{num_rgs_to_create}\n")

    workspace_event_status = "in_progress"
    if num_resources_created == num_resources_to_create:
        workspace_event_status = "success"
    else:
        logger.error("Some resources failed to create, please check logs")
        workspace_event_status = "failed"

    if phi_config.user is not None and ws_config.ws_schema is not None and ws_config.ws_schema.id_workspace is not None:
        # Log workspace start event
        log_workspace_event(
            user=phi_config.user,
            workspace_event=WorkspaceEvent(
                id_workspace=ws_config.ws_schema.id_workspace,
                event_type="start",
                event_status=workspace_event_status,
                event_data={
                    "target_env": target_env,
                    "target_infra": target_infra,
                    "target_group": target_group,
                    "target_name": target_name,
                    "target_type": target_type,
                    "dry_run": dry_run,
                    "auto_confirm": auto_confirm,
                    "force": force,
                },
            ),
        )


def stop_workspace(
    phi_config: PhiCliConfig,
    ws_config: WorkspaceConfig,
    target_env: Optional[str] = None,
    target_infra: Optional[InfraType] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
) -> None:
    """Stop a Phi Workspace. This is called from `phi ws down`"""
    if ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return

    # Set the local environment variables before processing configs
    ws_config.set_local_env()

    # Get resource groups to delete
    resource_groups_to_delete: List[InfraResources] = ws_config.get_resources(
        env=target_env,
        infra=target_infra,
        order="delete",
    )

    # Track number of resource groups deleted
    num_rgs_deleted = 0
    num_rgs_to_delete = len(resource_groups_to_delete)
    # Track number of resources deleted
    num_resources_deleted = 0
    num_resources_to_delete = 0

    if num_rgs_to_delete == 0:
        print_info("No resources to delete")
        return

    logger.debug(f"Deleting {num_rgs_to_delete} resource groups")
    for rg in resource_groups_to_delete:
        _num_resources_deleted, _num_resources_to_delete = rg.delete_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
        )
        if _num_resources_deleted > 0:
            num_rgs_deleted += 1
        num_resources_deleted += _num_resources_deleted
        num_resources_to_delete += _num_resources_to_delete
        logger.debug(f"Deleted {num_resources_deleted} resources in {num_rgs_deleted} resource groups")

    if dry_run:
        return

    if num_resources_deleted == 0:
        return

    print_heading(f"\n--**-- ResourceGroups deleted: {num_rgs_deleted}/{num_rgs_to_delete}\n")

    workspace_event_status = "in_progress"
    if num_resources_to_delete == num_resources_deleted:
        workspace_event_status = "success"
    else:
        logger.error("Some resources failed to delete, please check logs")
        workspace_event_status = "failed"

    if phi_config.user is not None and ws_config.ws_schema is not None and ws_config.ws_schema.id_workspace is not None:
        # Log workspace stop event
        log_workspace_event(
            user=phi_config.user,
            workspace_event=WorkspaceEvent(
                id_workspace=ws_config.ws_schema.id_workspace,
                event_type="stop",
                event_status=workspace_event_status,
                event_data={
                    "target_env": target_env,
                    "target_infra": target_infra,
                    "target_group": target_group,
                    "target_name": target_name,
                    "target_type": target_type,
                    "dry_run": dry_run,
                    "auto_confirm": auto_confirm,
                    "force": force,
                },
            ),
        )


def update_workspace(
    phi_config: PhiCliConfig,
    ws_config: WorkspaceConfig,
    target_env: Optional[str] = None,
    target_infra: Optional[InfraType] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
) -> None:
    """Update a Phi Workspace. This is called from `phi ws patch`"""
    if ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return

    # Set the local environment variables before processing configs
    ws_config.set_local_env()

    # Get resource groups to update
    resource_groups_to_update: List[InfraResources] = ws_config.get_resources(
        env=target_env,
        infra=target_infra,
        order="create",
    )
    # Track number of resource groups updated
    num_rgs_updated = 0
    num_rgs_to_update = len(resource_groups_to_update)
    # Track number of resources updated
    num_resources_updated = 0
    num_resources_to_update = 0

    if num_rgs_to_update == 0:
        print_info("No resources to update")
        return

    logger.debug(f"Updating {num_rgs_to_update} resource groups")
    for rg in resource_groups_to_update:
        _num_resources_updated, _num_resources_to_update = rg.update_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
        )
        if _num_resources_updated > 0:
            num_rgs_updated += 1
        num_resources_updated += _num_resources_updated
        num_resources_to_update += _num_resources_to_update
        logger.debug(f"Updated {num_resources_updated} resources in {num_rgs_updated} resource groups")

    if dry_run:
        return

    if num_resources_updated == 0:
        return

    print_heading(f"\n--**-- ResourceGroups updated: {num_rgs_updated}/{num_rgs_to_update}\n")

    workspace_event_status = "in_progress"
    if num_resources_updated == num_resources_to_update:
        workspace_event_status = "success"
    else:
        logger.error("Some resources failed to update, please check logs")
        workspace_event_status = "failed"

    if phi_config.user is not None and ws_config.ws_schema is not None and ws_config.ws_schema.id_workspace is not None:
        # Log workspace start event
        log_workspace_event(
            user=phi_config.user,
            workspace_event=WorkspaceEvent(
                id_workspace=ws_config.ws_schema.id_workspace,
                event_type="update",
                event_status=workspace_event_status,
                event_data={
                    "target_env": target_env,
                    "target_infra": target_infra,
                    "target_group": target_group,
                    "target_name": target_name,
                    "target_type": target_type,
                    "dry_run": dry_run,
                    "auto_confirm": auto_confirm,
                    "force": force,
                },
            ),
        )


def delete_workspace(phi_config: PhiCliConfig, ws_to_delete: Optional[List[Path]]) -> None:
    if ws_to_delete is None or len(ws_to_delete) == 0:
        print_heading("No workspaces to delete")
        return

    for ws_root in ws_to_delete:
        phi_config.delete_ws(ws_root_path=ws_root)


def set_workspace_as_active(ws_dir_name: Optional[str]) -> None:
    from phi.cli.operator import initialize_phi

    ######################################################
    ## 1. Validate Pre-requisites
    ######################################################
    ######################################################
    # 1.1 Check PhiConf is valid
    ######################################################
    phi_config: Optional[PhiCliConfig] = PhiCliConfig.from_saved_config()
    if not phi_config:
        # Phidata should be initialized before workspace setup
        init_success = initialize_phi()
        if not init_success:
            from phi.cli.console import log_phi_init_failed_msg

            log_phi_init_failed_msg()
            return
        phi_config = PhiCliConfig.from_saved_config()
        # If phi_config is still None, throw an error
        if not phi_config:
            raise Exception("Failed to initialize phi")

    ######################################################
    # 1.2 Check ws_root_path is valid
    ######################################################
    # By default, we assume this command is run from the workspace directory
    ws_root_path: Optional[Path] = None
    if ws_dir_name is None:
        # If the user does not provide a ws_name, that implies `phi set` is ran from
        # the workspace directory.
        ws_root_path = Path(".").resolve()
    else:
        # If the user provides a workspace name manually, we find the dir for that ws
        ws_config: Optional[WorkspaceConfig] = phi_config.get_ws_config_by_dir_name(ws_dir_name)
        if ws_config is None:
            logger.error(f"Could not find workspace {ws_dir_name}")
            return
        ws_root_path = ws_config.ws_root_path

    ws_dir_is_valid: bool = ws_root_path is not None and ws_root_path.exists() and ws_root_path.is_dir()
    if not ws_dir_is_valid:
        logger.error("Invalid workspace directory: {}".format(ws_root_path))
        return

    ######################################################
    # 1.3 Validate PhiWsData is available i.e. a workspace is available at this directory
    ######################################################
    logger.debug(f"Checking for a workspace at path: {ws_root_path}")
    active_ws_config: Optional[WorkspaceConfig] = phi_config.get_ws_config_by_path(ws_root_path)
    if active_ws_config is None:
        # This happens when the workspace is not yet setup
        print_info(f"Could not find a workspace at path: {ws_root_path}")
        print_info("If this workspace has not been setup, please run `phi ws setup` from the workspace directory")
        return

    print_heading(f"Setting workspace {active_ws_config.ws_root_path.stem} as active")
    # if load:
    #     try:
    #         active_ws_config.load()
    #     except Exception as e:
    #         logger.error("Could not load workspace config, please fix errors and try again")
    #         logger.error(e)
    #         return

    ######################################################
    # 1.4 Make api request if updating active workspace
    ######################################################
    logger.debug("Updating active workspace api")
    if phi_config.user is not None:
        ws_schema: Optional[WorkspaceSchema] = active_ws_config.ws_schema
        if ws_schema is None:
            logger.warning(f"Please setup {active_ws_config.ws_root_path.stem} by running `phi ws setup`")
        else:
            from phi.api.workspace import update_primary_workspace_for_user

            updated_workspace_schema = update_primary_workspace_for_user(
                user=phi_config.user,
                workspace=UpdatePrimaryWorkspace(
                    id_workspace=ws_schema.id_workspace,
                    ws_name=ws_schema.ws_name,
                ),
            )
            if updated_workspace_schema is not None:
                # Update the ws_schema for this workspace.
                phi_config.update_ws_config(
                    ws_root_path=active_ws_config.ws_root_path, ws_schema=updated_workspace_schema
                )

    ######################################################
    ## 2. Set workspace as active
    ######################################################
    phi_config.set_active_ws_dir(active_ws_config.ws_root_path)
    print_info("Active workspace updated")
    return



from pathlib import Path
from textwrap import dedent
from typing import Optional, Any, Dict, List, Union

from phi.aws.api_client import AwsApiClient
from phi.aws.resource.base import AwsResource
from phi.aws.resource.iam.role import IamRole
from phi.aws.resource.cloudformation.stack import CloudFormationStack
from phi.aws.resource.ec2.subnet import Subnet
from phi.aws.resource.eks.addon import EksAddon
from phi.cli.console import print_info
from phi.utils.log import logger


class EksCluster(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/eks.html
    """

    resource_type: Optional[str] = "EksCluster"
    service_name: str = "eks"

    # The unique name to give to your cluster.
    name: str
    # version: The desired Kubernetes version for your cluster.
    # If you don't specify a value here, the latest version available in Amazon EKS is used.
    version: Optional[str] = None

    # role: The IAM role that provides permissions for the Kubernetes control plane to make calls
    # to Amazon Web Services API operations on your behalf.
    # ARN for the EKS IAM role to use
    role_arn: Optional[str] = None
    # If role_arn is None, a default role is created if create_role is True
    create_role: bool = True
    # Provide IamRole to create or use default of role is None
    role: Optional[IamRole] = None
    # Name for the default role when role is None, use "name-role" if not provided
    role_name: Optional[str] = None
    # Provide a list of policy ARNs to attach to the role
    add_policy_arns: Optional[List[str]] = None

    # EKS VPC Configuration
    # resources_vpc_config: The VPC configuration that's used by the cluster control plane.
    # Amazon EKS VPC resources have specific requirements to work properly with Kubernetes.
    # You must specify at least two subnets. You can specify up to five security groups.
    resources_vpc_config: Optional[Dict[str, Any]] = None
    # If resources_vpc_config is None, a default CloudFormationStack is created if create_vpc_stack is True
    create_vpc_stack: bool = True
    # The CloudFormationStack to build resources_vpc_config if provided
    vpc_stack: Optional[CloudFormationStack] = None
    # If resources_vpc_config and vpc_stack are None
    # create a default CloudFormationStack using vpc_stack_name, use "name-vpc-stack" if vpc_stack_name is None
    vpc_stack_name: Optional[str] = None
    # Default VPC Stack Template URL
    vpc_stack_template_url: str = (
        "https://s3.us-west-2.amazonaws.com/amazon-eks/cloudformation/2020-10-29/amazon-eks-vpc-private-subnets.yaml"
    )
    use_public_subnets: bool = True
    use_private_subnets: bool = True
    subnet_az: Optional[Union[str, List[str]]] = None
    add_subnets: Optional[List[str]] = None
    add_security_groups: Optional[List[str]] = None
    endpoint_public_access: Optional[bool] = None
    endpoint_private_access: Optional[bool] = None
    public_access_cidrs: Optional[List[str]] = None

    # The Kubernetes network configuration for the cluster.
    kubernetes_network_config: Optional[Dict[str, str]] = None
    # Enable or disable exporting the Kubernetes control plane logs for your cluster to CloudWatch Logs.
    # By default, cluster control plane logs aren't exported to CloudWatch Logs.
    logging: Optional[Dict[str, List[dict]]] = None
    # Unique, case-sensitive identifier that you provide to ensure the idempotency of the request.
    client_request_token: Optional[str] = None
    # The metadata to apply to the cluster to assist with categorization and organization.
    # Each tag consists of a key and an optional value. You define both.
    tags: Optional[Dict[str, str]] = None
    # The encryption configuration for the cluster.
    encryption_config: Optional[List[Dict[str, Union[List[str], Dict[str, str]]]]] = None

    # EKS Addons
    addons: List[Union[str, EksAddon]] = ["aws-ebs-csi-driver", "aws-efs-csi-driver", "vpc-cni", "coredns"]

    # Kubeconfig
    # If True, updates the kubeconfig on create/delete
    # Use manage_kubeconfig = False when using a separate EksKubeconfig resource
    manage_kubeconfig: bool = True
    # The kubeconfig_path to update
    kubeconfig_path: Path = Path.home().joinpath(".kube").joinpath("config").resolve()
    # Optional: cluster_name to use in kubeconfig, defaults to self.name
    kubeconfig_cluster_name: Optional[str] = None
    # Optional: cluster_user to use in kubeconfig, defaults to self.name
    kubeconfig_cluster_user: Optional[str] = None
    # Optional: cluster_context to use in kubeconfig, defaults to self.name
    kubeconfig_cluster_context: Optional[str] = None
    # Optional: role to assume when signing the token
    kubeconfig_role: Optional[IamRole] = None
    # Optional: role arn to assume when signing the token
    kubeconfig_role_arn: Optional[str] = None

    # provided by api on create
    created_at: Optional[str] = None
    cluster_status: Optional[str] = None

    # bump the wait time for Eks to 30 seconds
    waiter_delay: int = 30

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the EksCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Get IamRoleArn
        eks_iam_role_arn = self.role_arn
        if eks_iam_role_arn is None and self.create_role:
            # Create the IamRole and get eks_iam_role_arn
            eks_iam_role = self.get_eks_iam_role()
            try:
                eks_iam_role.create(aws_client)
                eks_iam_role_arn = eks_iam_role.read(aws_client).arn
                print_info(f"ARN for {eks_iam_role.name}: {eks_iam_role_arn}")
            except Exception as e:
                logger.error("IamRole creation failed, please fix and try again")
                logger.error(e)
                return False
        if eks_iam_role_arn is None:
            logger.error("IamRole ARN not available, please fix and try again")
            return False

        # Step 2: Get the VPC config
        resources_vpc_config = self.resources_vpc_config
        if resources_vpc_config is None and self.create_vpc_stack:
            print_info("Creating default vpc stack as no resources_vpc_config provided")
            # Create the CloudFormationStack and get resources_vpc_config
            vpc_stack = self.get_vpc_stack()
            try:
                vpc_stack.create(aws_client)
                resources_vpc_config = self.get_eks_resources_vpc_config(aws_client, vpc_stack)
            except Exception as e:
                logger.error("Stack creation failed, please fix and try again")
                logger.error(e)
                return False
        if resources_vpc_config is None:
            logger.error("VPC configuration not available, please fix and try again")
            return False

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.version:
            not_null_args["version"] = self.version
        if self.kubernetes_network_config:
            not_null_args["kubernetesNetworkConfig"] = self.kubernetes_network_config
        if self.logging:
            not_null_args["logging"] = self.logging
        if self.client_request_token:
            not_null_args["clientRequestToken"] = self.client_request_token
        if self.tags:
            not_null_args["tags"] = self.tags
        if self.encryption_config:
            not_null_args["encryptionConfig"] = self.encryption_config

        # Step 3: Create EksCluster
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_cluster(
                name=self.name,
                roleArn=eks_iam_role_arn,
                resourcesVpcConfig=resources_vpc_config,
                **not_null_args,
            )
            logger.debug(f"EksCluster: {create_response}")
            cluster_dict = create_response.get("cluster", {})

            # Validate Cluster creation
            self.created_at = cluster_dict.get("createdAt", None)
            self.cluster_status = cluster_dict.get("status", None)
            logger.debug(f"created_at: {self.created_at}")
            logger.debug(f"cluster_status: {self.cluster_status}")
            if self.created_at is not None:
                print_info(f"EksCluster created: {self.name}")
                self.active_resource = create_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # Wait for Cluster to be created
        if self.wait_for_create:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be active.")
                waiter = self.get_service_client(aws_client).get_waiter("cluster_active")
                waiter.wait(
                    name=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)

        # Add addons
        if self.addons is not None:
            addons_created: List[EksAddon] = []
            for _addon in self.addons:
                addon_to_create: Optional[EksAddon] = None
                if isinstance(_addon, EksAddon):
                    addon_to_create = _addon
                elif isinstance(_addon, str):
                    addon_to_create = EksAddon(name=_addon, cluster_name=self.name)

                if addon_to_create is not None:
                    addon_success = addon_to_create._create(aws_client)  # type: ignore
                    if addon_success:
                        addons_created.append(addon_to_create)

            # Wait for Addons to be created
            if self.wait_for_create:
                for addon in addons_created:
                    addon.post_create(aws_client)

        # Update kubeconfig if needed
        if self.manage_kubeconfig:
            self.write_kubeconfig(aws_client=aws_client)
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the EksCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            describe_response = service_client.describe_cluster(name=self.name)
            # logger.debug(f"EksCluster: {describe_response}")
            cluster_dict = describe_response.get("cluster", {})

            self.created_at = cluster_dict.get("createdAt", None)
            self.cluster_status = cluster_dict.get("status", None)
            logger.debug(f"EksCluster created_at: {self.created_at}")
            logger.debug(f"EksCluster status: {self.cluster_status}")
            if self.created_at is not None:
                logger.debug(f"EksCluster found: {self.name}")
                self.active_resource = describe_response
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the EksCluster
        Deletes the Amazon EKS cluster control plane.
        If you have active services in your cluster that are associated with a load balancer,
        you must delete those services before deleting the cluster so that the load balancers
        are deleted properly. Otherwise, you can have orphaned resources in your VPC
        that prevent you from being able to delete the VPC.

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Delete the IamRole
        if self.role_arn is None and self.create_role:
            eks_iam_role = self.get_eks_iam_role()
            try:
                eks_iam_role.delete(aws_client)
            except Exception as e:
                logger.error("IamRole deletion failed, please try again or delete manually")
                logger.error(e)

        # Step 2: Delete the CloudFormationStack if needed
        if self.resources_vpc_config is None and self.create_vpc_stack:
            vpc_stack = self.get_vpc_stack()
            try:
                vpc_stack.delete(aws_client)
            except Exception as e:
                logger.error("Stack deletion failed, please try again or delete manually")
                logger.error(e)

        # Step 3: Delete the EksCluster
        service_client = self.get_service_client(aws_client)
        self.active_resource = None
        try:
            delete_response = service_client.delete_cluster(name=self.name)
            logger.debug(f"EksCluster: {delete_response}")
            return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        # Wait for Cluster to be deleted
        if self.wait_for_delete:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be deleted.")
                waiter = self.get_service_client(aws_client).get_waiter("cluster_deleted")
                waiter.wait(
                    name=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)

        # Update kubeconfig if needed
        if self.manage_kubeconfig:
            return self.clean_kubeconfig(aws_client=aws_client)
        return True

    def get_eks_iam_role(self) -> IamRole:
        if self.role is not None:
            return self.role

        policy_arns = ["arn:aws:iam::aws:policy/AmazonEKSClusterPolicy"]
        if self.add_policy_arns is not None and isinstance(self.add_policy_arns, list):
            policy_arns.extend(self.add_policy_arns)

        return IamRole(
            name=self.role_name or f"{self.name}-role",
            assume_role_policy_document=dedent(
                """\
            {
              "Version": "2012-10-17",
              "Statement": [
                {
                  "Effect": "Allow",
                  "Principal": {
                    "Service": "eks.amazonaws.com"
                  },
                  "Action": "sts:AssumeRole"
                }
              ]
            }
            """
            ),
            policy_arns=policy_arns,
        )

    def get_vpc_stack(self) -> CloudFormationStack:
        if self.vpc_stack is not None:
            return self.vpc_stack
        return CloudFormationStack(
            name=self.vpc_stack_name or f"{self.name}-vpc",
            template_url=self.vpc_stack_template_url,
            skip_create=self.skip_create,
            skip_delete=self.skip_delete,
            wait_for_create=self.wait_for_create,
            wait_for_delete=self.wait_for_delete,
        )

    def get_subnets(self, aws_client: AwsApiClient, vpc_stack: Optional[CloudFormationStack] = None) -> List[str]:
        subnet_ids: List[str] = []

        # Option 1: Get subnets from the resources_vpc_config provided by the user
        if self.resources_vpc_config is not None and "subnetIds" in self.resources_vpc_config:
            subnet_ids = self.resources_vpc_config["subnetIds"]
            if not isinstance(subnet_ids, list):
                raise TypeError(f"resources_vpc_config.subnetIds must be a list of strings, not {type(subnet_ids)}")
            return subnet_ids

        # Option 2: Get subnets from the cloudformation VPC stack
        if vpc_stack is None:
            vpc_stack = self.get_vpc_stack()

        if self.use_public_subnets:
            public_subnets: Optional[List[str]] = vpc_stack.get_public_subnets(aws_client)
            if public_subnets is not None:
                subnet_ids.extend(public_subnets)

        if self.use_private_subnets:
            private_subnets: Optional[List[str]] = vpc_stack.get_private_subnets(aws_client)
            if private_subnets is not None:
                subnet_ids.extend(private_subnets)

        if self.subnet_az is not None:
            azs_filter = []
            if isinstance(self.subnet_az, str):
                azs_filter.append(self.subnet_az)
            elif isinstance(self.subnet_az, list):
                azs_filter.extend(self.subnet_az)

            subnet_ids = [
                subnet_id
                for subnet_id in subnet_ids
                if Subnet(name=subnet_id).get_availability_zone(aws_client=aws_client) in azs_filter
            ]
        return subnet_ids

    def get_eks_resources_vpc_config(
        self, aws_client: AwsApiClient, vpc_stack: CloudFormationStack
    ) -> Dict[str, List[Any]]:
        if self.resources_vpc_config is not None:
            return self.resources_vpc_config

        # Build resources_vpc_config using vpc_stack
        # get the VPC physical_resource_id
        # vpc_stack_resource = vpc_stack.get_stack_resource(aws_client, "VPC")
        # vpc_physical_resource_id = (
        #     vpc_stack_resource.physical_resource_id
        #     if vpc_stack_resource is not None
        #     else None
        # )
        # # logger.debug(f"vpc_physical_resource_id: {vpc_physical_resource_id}")

        # get the ControlPlaneSecurityGroup physical_resource_id
        sg_stack_resource = vpc_stack.get_stack_resource(aws_client, "ControlPlaneSecurityGroup")
        sg_physical_resource_id = sg_stack_resource.physical_resource_id if sg_stack_resource is not None else None
        security_group_ids = [sg_physical_resource_id] if sg_physical_resource_id is not None else []
        if self.add_security_groups is not None and isinstance(self.add_security_groups, list):
            security_group_ids.extend(self.add_security_groups)
        logger.debug(f"security_group_ids: {security_group_ids}")

        subnet_ids: List[str] = self.get_subnets(aws_client, vpc_stack)
        if self.add_subnets is not None and isinstance(self.add_subnets, list):
            subnet_ids.extend(self.add_subnets)
        logger.debug(f"subnet_ids: {subnet_ids}")

        resources_vpc_config: Dict[str, Any] = {
            "subnetIds": subnet_ids,
            "securityGroupIds": security_group_ids,
        }

        if self.endpoint_public_access is not None:
            resources_vpc_config["endpointPublicAccess"] = self.endpoint_public_access
        if self.endpoint_private_access is not None:
            resources_vpc_config["endpointPrivateAccess"] = self.endpoint_private_access
        if self.public_access_cidrs is not None:
            resources_vpc_config["publicAccessCidrs"] = self.public_access_cidrs

        return resources_vpc_config

    def get_subnets_in_order(self, aws_client: AwsApiClient) -> List[str]:
        """
        Returns the subnet_ids in the following order:
            - User provided subnets
            - Private subnets from the VPC stack
            - Public subnets from the VPC stack
        """
        # Option 1: Get subnets from the resources_vpc_config provided by the user
        if self.resources_vpc_config is not None and "subnetIds" in self.resources_vpc_config:
            subnet_ids = self.resources_vpc_config["subnetIds"]
            if not isinstance(subnet_ids, list):
                raise TypeError(f"resources_vpc_config.subnetIds must be a list of strings, not {type(subnet_ids)}")
            return subnet_ids

        # Option 2: Get private subnets from the VPC stack
        vpc_stack = self.get_vpc_stack()
        if self.use_private_subnets:
            private_subnets: Optional[List[str]] = vpc_stack.get_private_subnets(aws_client)
            if private_subnets is not None:
                return private_subnets

        # Option 3: Get public subnets from the VPC stack
        if self.use_public_subnets:
            public_subnets: Optional[List[str]] = vpc_stack.get_public_subnets(aws_client)
            if public_subnets is not None:
                return public_subnets
        return []

    def get_kubeconfig_cluster_name(self) -> str:
        return self.kubeconfig_cluster_name or self.name

    def get_kubeconfig_user_name(self) -> str:
        return self.kubeconfig_cluster_user or self.name

    def get_kubeconfig_context_name(self) -> str:
        return self.kubeconfig_cluster_context or self.name

    def write_kubeconfig(self, aws_client: AwsApiClient) -> bool:
        # Step 1: Get the EksCluster to generate the kubeconfig for
        eks_cluster = self._read(aws_client)
        if eks_cluster is None:
            logger.warning(f"EKSCluster not available: {self.name}")
            return False

        # Step 2: Get EksCluster cert, endpoint & arn
        try:
            cluster_cert = eks_cluster.get("cluster", {}).get("certificateAuthority", {}).get("data", None)
            logger.debug(f"cluster_cert: {cluster_cert}")

            cluster_endpoint = eks_cluster.get("cluster", {}).get("endpoint", None)
            logger.debug(f"cluster_endpoint: {cluster_endpoint}")

            cluster_arn = eks_cluster.get("cluster", {}).get("arn", None)
            logger.debug(f"cluster_arn: {cluster_arn}")
        except Exception as e:
            logger.error("Cannot read EKSCluster")
            logger.error(e)
            return False

        # from phi.k8s.enums.api_version import ApiVersion
        # from phi.k8s.resource.kubeconfig import (
        #     Kubeconfig,
        #     KubeconfigCluster,
        #     KubeconfigClusterConfig,
        #     KubeconfigContext,
        #     KubeconfigContextSpec,
        #     KubeconfigUser,
        # )
        #
        # # Step 3: Build Kubeconfig components
        # # 3.1 Build KubeconfigCluster config
        # new_cluster = KubeconfigCluster(
        #     name=self.get_kubeconfig_cluster_name(),
        #     cluster=KubeconfigClusterConfig(
        #         server=str(cluster_endpoint),
        #         certificate_authority_data=str(cluster_cert),
        #     ),
        # )
        #
        # # 3.2 Build KubeconfigUser config
        # new_user_exec_args = ["eks", "get-token", "--cluster-name", self.name]
        # if aws_client.aws_region is not None:
        #     new_user_exec_args.extend(["--region", aws_client.aws_region])
        # # Assume the role if the role_arn is provided
        # if self.kubeconfig_role_arn is not None:
        #     new_user_exec_args.extend(["--role-arn", self.kubeconfig_role_arn])
        # # Otherwise if role is provided, use that to get the role arn
        # elif self.kubeconfig_role is not None:
        #     _arn = self.kubeconfig_role.get_arn(aws_client=aws_client)
        #     if _arn is not None:
        #         new_user_exec_args.extend(["--role-arn", _arn])
        #
        # new_user_exec: Dict[str, Any] = {
        #     "apiVersion": ApiVersion.CLIENT_AUTHENTICATION_V1BETA1.value,
        #     "command": "aws",
        #     "args": new_user_exec_args,
        # }
        # if aws_client.aws_profile is not None:
        #     new_user_exec["env"] = [{"name": "AWS_PROFILE", "value": aws_client.aws_profile}]
        #
        # new_user = KubeconfigUser(
        #     name=self.get_kubeconfig_user_name(),
        #     user={"exec": new_user_exec},
        # )
        #
        # # 3.3 Build KubeconfigContext config
        # new_context = KubeconfigContext(
        #     name=self.get_kubeconfig_context_name(),
        #     context=KubeconfigContextSpec(
        #         cluster=new_cluster.name,
        #         user=new_user.name,
        #     ),
        # )
        # current_context = new_context.name
        # cluster_config: KubeconfigCluster
        #
        # # Step 4: Get existing Kubeconfig
        # kubeconfig_path = self.kubeconfig_path
        # if kubeconfig_path is None:
        #     logger.error(f"kubeconfig_path is None")
        #     return False
        #
        # kubeconfig: Optional[Any] = Kubeconfig.read_from_file(kubeconfig_path)
        #
        # # Step 5: Parse through the existing config to determine if
        # # an update is required. By the end of this logic
        # # if write_kubeconfig = False then no changes to kubeconfig are needed
        # # if write_kubeconfig = True then we should write the kubeconfig file
        # write_kubeconfig = False
        #
        # # Kubeconfig exists and is valid
        # if kubeconfig is not None and isinstance(kubeconfig, Kubeconfig):
        #     # Update Kubeconfig.clusters:
        #     # If a cluster with the same name exists in Kubeconfig.clusters
        #     #   - check if server and cert values match, if not, remove the existing cluster
        #     #   and add the new cluster config. Mark cluster_config_exists = True
        #     # If a cluster with the same name does not exist in Kubeconfig.clusters
        #     #   - add the new cluster config
        #     cluster_config_exists = False
        #     for idx, _cluster in enumerate(kubeconfig.clusters, start=0):
        #         if _cluster.name == new_cluster.name:
        #             cluster_config_exists = True
        #             if (
        #                 _cluster.cluster.server != new_cluster.cluster.server
        #                 or _cluster.cluster.certificate_authority_data
        #                 != new_cluster.cluster.certificate_authority_data
        #             ):
        #                 logger.debug("Kubeconfig.cluster mismatch, updating cluster config")
        #                 removed_cluster_config = kubeconfig.clusters.pop(idx)
        #                 # logger.debug(
        #                 #     f"removed_cluster_config: {removed_cluster_config}"
        #                 # )
        #                 kubeconfig.clusters.append(new_cluster)
        #                 write_kubeconfig = True
        #     if not cluster_config_exists:
        #         logger.debug("Adding Kubeconfig.cluster")
        #         kubeconfig.clusters.append(new_cluster)
        #         write_kubeconfig = True
        #
        #     # Update Kubeconfig.users:
        #     # If a user with the same name exists in Kubeconfig.users -
        #     #   check if user spec matches, if not, remove the existing user
        #     #   and add the new user config. Mark user_config_exists = True
        #     # If a user with the same name does not exist in Kubeconfig.users -
        #     #   add the new user config
        #     user_config_exists = False
        #     for idx, _user in enumerate(kubeconfig.users, start=0):
        #         if _user.name == new_user.name:
        #             user_config_exists = True
        #             if _user.user != new_user.user:
        #                 logger.debug("Kubeconfig.user mismatch, updating user config")
        #                 removed_user_config = kubeconfig.users.pop(idx)
        #                 # logger.debug(f"removed_user_config: {removed_user_config}")
        #                 kubeconfig.users.append(new_user)
        #                 write_kubeconfig = True
        #     if not user_config_exists:
        #         logger.debug("Adding Kubeconfig.user")
        #         kubeconfig.users.append(new_user)
        #         write_kubeconfig = True
        #
        #     # Update Kubeconfig.contexts:
        #     # If a context with the same name exists in Kubeconfig.contexts -
        #     #   check if context spec matches, if not, remove the existing context
        #     #   and add the new context. Mark context_config_exists = True
        #     # If a context with the same name does not exist in Kubeconfig.contexts -
        #     #   add the new context config
        #     context_config_exists = False
        #     for idx, _context in enumerate(kubeconfig.contexts, start=0):
        #         if _context.name == new_context.name:
        #             context_config_exists = True
        #             if _context.context != new_context.context:
        #                 logger.debug("Kubeconfig.context mismatch, updating context config")
        #                 removed_context_config = kubeconfig.contexts.pop(idx)
        #                 # logger.debug(
        #                 #     f"removed_context_config: {removed_context_config}"
        #                 # )
        #                 kubeconfig.contexts.append(new_context)
        #                 write_kubeconfig = True
        #     if not context_config_exists:
        #         logger.debug("Adding Kubeconfig.context")
        #         kubeconfig.contexts.append(new_context)
        #         write_kubeconfig = True
        #
        #     if kubeconfig.current_context is None or kubeconfig.current_context != current_context:
        #         logger.debug("Updating Kubeconfig.current_context")
        #         kubeconfig.current_context = current_context
        #         write_kubeconfig = True
        # else:
        #     # Kubeconfig does not exist or is not valid
        #     # Create a new Kubeconfig
        #     logger.info(f"Creating new Kubeconfig")
        #     kubeconfig = Kubeconfig(
        #         clusters=[new_cluster],
        #         users=[new_user],
        #         contexts=[new_context],
        #         current_context=current_context,
        #     )
        #     write_kubeconfig = True
        #
        # # if kubeconfig:
        # #     logger.debug("Kubeconfig:\n{}".format(kubeconfig.json(exclude_none=True, by_alias=True, indent=4)))
        #
        # # Step 5: Write Kubeconfig if an update is made
        # if write_kubeconfig:
        #     return kubeconfig.write_to_file(kubeconfig_path)
        # else:
        #     logger.info("Kubeconfig up-to-date")
        return True

    def clean_kubeconfig(self, aws_client: AwsApiClient) -> bool:
        logger.debug(f"TO_DO: Cleaning kubeconfig at {str(self.kubeconfig_path)}")
        return True



from typing import Optional, Any, List

try:
    from sqlalchemy.dialects import sqlite
    from sqlalchemy.engine import create_engine, Engine
    from sqlalchemy.engine.row import Row
    from sqlalchemy.inspection import inspect
    from sqlalchemy.orm import Session, sessionmaker
    from sqlalchemy.schema import MetaData, Table, Column
    from sqlalchemy.sql.expression import select
    from sqlalchemy.types import String
except ImportError:
    raise ImportError("`sqlalchemy` not installed")

from sqlite3 import OperationalError

from phi.conversation.schemas import ConversationRow
from phi.conversation.storage.base import ConversationStorage
from phi.utils.dttm import current_datetime
from phi.utils.log import logger


class SqlConversationStorage(ConversationStorage):
    def __init__(
        self,
        table_name: str,
        db_url: Optional[str] = None,
        db_file: Optional[str] = None,
        db_engine: Optional[Engine] = None,
    ):
        """
        This class provides conversation storage using a sqllite database.

        The following order is used to determine the database connection:
            1. Use the db_engine if provided
            2. Use the db_url
            3. Use the db_file
            4. Create a new in-memory database

        :param table_name: The name of the table to store conversations in.
        :param db_url: The database URL to connect to.
        :param db_file: The database file to connect to.
        :param db_engine: The database engine to use.
        """
        _engine: Optional[Engine] = db_engine
        if _engine is None and db_url is not None:
            _engine = create_engine(db_url)
        elif _engine is None and db_file is not None:
            _engine = create_engine(f"sqlite:///{db_file}")
        else:
            _engine = create_engine("sqlite://")

        if _engine is None:
            raise ValueError("Must provide either db_url, db_file or db_engine")

        # Database attributes
        self.table_name: str = table_name
        self.db_url: Optional[str] = db_url
        self.db_engine: Engine = _engine
        self.metadata: MetaData = MetaData()

        # Database session
        self.Session: sessionmaker[Session] = sessionmaker(bind=self.db_engine)

        # Database table for storage
        self.table: Table = self.get_table()

    def get_table(self) -> Table:
        return Table(
            self.table_name,
            self.metadata,
            # Database ID/Primary key for this conversation.
            Column("id", String, primary_key=True),
            # Conversation name
            Column("name", String),
            # Name and type of user participating in this conversation.
            Column("user_name", String),
            Column("user_type", String),
            # True if this conversation is active.
            Column("is_active", sqlite.BOOLEAN, default=True),
            # -*- LLM data (name, model, etc.)
            Column("llm", sqlite.JSON),
            # -*- Conversation memory
            Column("memory", sqlite.JSON),
            # Metadata associated with this conversation.
            Column("meta_data", sqlite.JSON),
            # Extra data associated with this conversation.
            Column("extra_data", sqlite.JSON),
            # The timestamp of when this conversation was created.
            Column("created_at", sqlite.DATETIME, default=current_datetime()),
            # The timestamp of when this conversation was last updated.
            Column("updated_at", sqlite.DATETIME, onupdate=current_datetime()),
            extend_existing=True,
            sqlite_autoincrement=True,
        )

    def table_exists(self) -> bool:
        logger.debug(f"Checking if table exists: {self.table.name}")
        try:
            return inspect(self.db_engine).has_table(self.table.name)
        except Exception as e:
            logger.error(e)
            return False

    def create(self) -> None:
        if not self.table_exists():
            logger.debug(f"Creating table: {self.table.name}")
            self.table.create(self.db_engine)

    def _read(self, session: Session, conversation_id: str) -> Optional[Row[Any]]:
        stmt = select(self.table).where(self.table.c.id == conversation_id)
        try:
            return session.execute(stmt).first()
        except OperationalError:
            # Create table if it does not exist
            self.create()
        except Exception as e:
            logger.warning(e)
        return None

    def read(self, conversation_id: str) -> Optional[ConversationRow]:
        with self.Session() as sess:
            existing_row: Optional[Row[Any]] = self._read(session=sess, conversation_id=conversation_id)
            return ConversationRow.model_validate(existing_row) if existing_row is not None else None

    def get_all_conversation_ids(self, user_name: Optional[str] = None) -> List[str]:
        conversation_ids: List[str] = []
        try:
            with self.Session() as sess:
                # get all conversation ids for this user
                stmt = select(self.table)
                if user_name is not None:
                    stmt = stmt.where(self.table.c.user_name == user_name)
                # order by created_at desc
                stmt = stmt.order_by(self.table.c.created_at.desc())
                # execute query
                rows = sess.execute(stmt).fetchall()
                for row in rows:
                    if row is not None and row.id is not None:
                        conversation_ids.append(row.id)
        except OperationalError:
            logger.debug(f"Table does not exist: {self.table.name}")
            pass
        return conversation_ids

    def get_all_conversations(self, user_name: Optional[str] = None) -> List[ConversationRow]:
        conversations: List[ConversationRow] = []
        try:
            with self.Session() as sess:
                # get all conversation ids for this user
                stmt = select(self.table)
                if user_name is not None:
                    stmt = stmt.where(self.table.c.user_name == user_name)
                # order by created_at desc
                stmt = stmt.order_by(self.table.c.created_at.desc())
                # execute query
                rows = sess.execute(stmt).fetchall()
                for row in rows:
                    if row.id is not None:
                        conversations.append(ConversationRow.model_validate(row))
        except OperationalError:
            logger.debug(f"Table does not exist: {self.table.name}")
            pass
        return conversations

    def upsert(self, conversation: ConversationRow) -> Optional[ConversationRow]:
        """
        Create a new conversation if it does not exist, otherwise update the existing conversation.
        """
        with self.Session() as sess:
            # Create an insert statement
            stmt = sqlite.insert(self.table).values(
                id=conversation.id,
                name=conversation.name,
                user_name=conversation.user_name,
                user_type=conversation.user_type,
                is_active=conversation.is_active,
                llm=conversation.llm,
                memory=conversation.memory,
                meta_data=conversation.meta_data,
                extra_data=conversation.extra_data,
            )

            # Define the upsert if the id already exists
            # See: https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#insert-on-conflict-upsert
            stmt = stmt.on_conflict_do_update(
                index_elements=["id"],
                set_=dict(
                    name=conversation.name,
                    user_name=conversation.user_name,
                    user_type=conversation.user_type,
                    is_active=conversation.is_active,
                    llm=conversation.llm,
                    memory=conversation.memory,
                    meta_data=conversation.meta_data,
                    extra_data=conversation.extra_data,
                ),  # The updated value for each column
            )

            try:
                sess.execute(stmt)
            except OperationalError:
                # Create table if it does not exist
                self.create()
                sess.execute(stmt)
        return self.read(conversation_id=conversation.id)

    def end(self, conversation_id: str) -> None:
        with self.Session() as sess:
            # Check if conversation exists in the database
            existing_row: Optional[Row[Any]] = self._read(session=sess, conversation_id=conversation_id)
            # If conversation exists, set is_active to False
            if existing_row is not None:
                stmt = self.table.update().where(self.table.c.id == conversation_id).values(is_active=False)
                sess.execute(stmt)
            sess.commit()

    def delete(self) -> None:
        if self.table_exists():
            logger.debug(f"Deleting table: {self.table_name}")
            self.table.drop(self.db_engine)



from typing import Any, Dict, Optional

from pydantic import BaseModel


class MonitorEventSchema(BaseModel):
    event_type: str
    event_status: str
    object_name: str
    event_data: Optional[Dict[str, Any]] = None
    object_data: Optional[Dict[str, Any]] = None


class MonitorResponseSchema(BaseModel):
    id_monitor: Optional[int] = None
    id_event: Optional[int] = None



from phi.base import PhiBase
from phi.k8s.resource.base import K8sResource, K8sObject
from phi.utils.log import logger


class CreateK8sObject(PhiBase):
    def _create(self) -> K8sObject:
        raise NotImplementedError

    def create(self) -> K8sObject:
        _resource = self._create()
        if _resource is None:
            raise ValueError(f"Failed to create resource: {self.__class__.__name__}")

        resource_fields = _resource.model_dump(exclude_defaults=True)
        base_fields = self.model_dump(exclude_defaults=True)

        # Get fields that are set for the base class but not the resource class
        diff_fields = {k: v for k, v in base_fields.items() if k not in resource_fields}

        updated_resource = _resource.model_copy(update=diff_fields)
        # logger.debug(f"Created resource: {updated_resource.__class__.__name__}: {updated_resource.model_dump()}")

        return updated_resource


class CreateK8sResource(PhiBase):
    def _create(self) -> K8sResource:
        raise NotImplementedError

    def create(self) -> K8sResource:
        _resource = self._create()
        # logger.debug(f"Created resource: {self.__class__.__name__}")
        if _resource is None:
            raise ValueError(f"Failed to create resource: {self.__class__.__name__}")

        resource_fields = _resource.model_dump(exclude_defaults=True)
        base_fields = self.model_dump(exclude_defaults=True)

        # Get fields that are set for the base class but not the resource class
        diff_fields = {k: v for k, v in base_fields.items() if k not in resource_fields}

        updated_resource = _resource.model_copy(update=diff_fields)
        # logger.debug(f"Created resource: {updated_resource.__class__.__name__}: {updated_resource.model_dump()}")

        logger.debug(f"Created: {updated_resource.__class__.__name__} | {updated_resource.get_resource_name()}")
        return updated_resource
